__author__ = 'Bohdan Mushkevych'

from subprocess import PIPE
import sys

import psutil
from psutil import TimeoutExpired

from launch import get_python, PROJECT_ROOT, PROCESS_STARTER
from synergy.system.utils import remove_pid_file, get_pid_filename
from synergy.conf import settings


def get_process_pid(process_name):
    """ check for process' pid file and returns pid from there """
    try:
        pid_filename = get_pid_filename(process_name)
        with open(pid_filename, mode='r') as pid_file:
            pid = int(pid_file.read().strip())
    except IOError:
        pid = None
    return pid


def kill_process(process_name):
    """ method is called to kill a running process """
    try:
        sys.stdout.write('killing: {0} {{ \n'.format(process_name))
        pid = get_process_pid(process_name)
        if pid is not None and psutil.pid_exists(int(pid)):
            p = psutil.Process(pid)
            p.kill()
            p.wait()
            remove_pid_file(process_name)
    except Exception as e:
        sys.stderr.write('Exception on killing {0} : {1} \n'.format(process_name, e))
    finally:
        sys.stdout.write('}')


def start_process(process_name, *args):
    try:
        sys.stdout.write('starting: {0} {{ \n'.format(process_name))
        cmd = [get_python(), PROJECT_ROOT + '/' + PROCESS_STARTER, process_name]
        if not args:
            # this blocks triggers when args is either None or an empty list
            pass
        else:
            cmd.extend(*args)

        p = psutil.Popen(cmd,
                         close_fds=True,
                         cwd=settings.settings['process_cwd'],
                         stdin=PIPE,
                         stdout=PIPE,
                         stderr=PIPE)
        sys.stdout.write('Started {0} with pid = {1} \n'.format(process_name, p.pid))
    except Exception as e:
        sys.stderr.write('Exception on starting {0} : {1} \n'.format(process_name, e))
    finally:
        sys.stdout.write('}')


def poll_process(process_name):
    """ between killing a process and its actual termination lies poorly documented requirement -
        <purging process' io pipes and reading exit status>.
        this can be done either by os.wait() or process.wait()
        :return True if the process is alive and OK and False is the process was terminated """
    try:
        pid = get_process_pid(process_name)
        if pid is None:
            sys.stdout.write('PID file was not found. Process {0} is likely terminated.\n'.format(process_name))
            return False

        p = psutil.Process(pid)
        return_code = p.wait(timeout=0.01)

        if return_code is None:
            # process is already terminated
            sys.stdout.write('Process {0} is terminated \n'.format(process_name))
            return False
        else:
            # process is terminated; possibly by OS
            sys.stdout.write('Process {0} got terminated \n'.format(process_name))
            return False
    except TimeoutExpired:
        sys.stdout.write('Process {0} is alive and OK \n'.format(process_name))
        return True



__author__ = 'Bohdan Mushkevych'

from odm.fields import BooleanField, StringField, DictField, ListField, NestedDocumentField
from odm.document import BaseDocument

from synergy.db.model.job import Job
from synergy.db.model.managed_process_entry import ManagedProcessEntry
from synergy.db.model.freerun_process_entry import FreerunProcessEntry
from synergy.db.model.timetable_tree_entry import TimetableTreeEntry


FIELD_IS_ALIVE = 'is_alive'                 # actual state of the trigger
FIELD_NEXT_RUN_IN = 'next_run_in'           # duration until the next trigger event in HH:MM:SS format
FIELD_NEXT_TIMEPERIOD = 'next_timeperiod'   # Synergy timeperiod format
FIELD_DEPENDANT_TREES = 'dependant_trees'
FIELD_SORTED_PROCESS_NAMES = 'sorted_process_names'     # process names sorted by their time_qualifier
FIELD_REPROCESSING_QUEUE = 'reprocessing_queue'

FIELD_NUMBER_OF_CHILDREN = 'number_of_children'
FIELD_PROCESSES = 'processes'
FIELD_TIME_QUALIFIER = 'time_qualifier'
FIELD_CHILDREN = 'children'
FIELD_NODE = 'node'


class RestFreerunSchedulerEntry(FreerunProcessEntry):
    is_alive = BooleanField(FIELD_IS_ALIVE)
    next_run_in = StringField(FIELD_NEXT_RUN_IN)


class RestManagedSchedulerEntry(ManagedProcessEntry):
    is_alive = BooleanField(FIELD_IS_ALIVE)
    next_run_in = StringField(FIELD_NEXT_RUN_IN)
    next_timeperiod = StringField(FIELD_NEXT_TIMEPERIOD)
    reprocessing_queue = ListField(FIELD_REPROCESSING_QUEUE)


class RestTimetableTree(TimetableTreeEntry):
    dependant_trees = ListField(FIELD_DEPENDANT_TREES)
    sorted_process_names = ListField(FIELD_SORTED_PROCESS_NAMES)
    processes = DictField(FIELD_PROCESSES)


class RestJob(Job):
    time_qualifier = StringField(FIELD_TIME_QUALIFIER)
    number_of_children = StringField(FIELD_NUMBER_OF_CHILDREN)


class RestTimetableTreeNode(BaseDocument):
    node = NestedDocumentField(FIELD_NODE, RestJob, null=True)
    children = DictField(FIELD_CHILDREN)



PROCESS_BASH_DRIVER = 'BashDriver'

TOKEN_BASH_DRIVER = 'bash'

# common arguments for remote command execution
ARGUMENT_CMD_PATH = 'cmd_path'
ARGUMENT_CMD_FILE = 'cmd_file'
ARGUMENT_CMD_ARGS = 'cmd_args'
ARGUMENT_CMD_HOST = 'cmd_host'



# Scheduler constants
LAG_5_MINUTES = 5  # lag in minutes between finish of the timeperiod and beginning of its processing

PROCESS_SCHEDULER = 'Scheduler'
PROCESS_GC = 'GarbageCollector'
PROCESS_MX = 'MX'

TOKEN_SCHEDULER = 'scheduler'
TOKEN_GC = 'gc'
TOKEN_WERKZEUG = 'werkzeug'

# Job is run once per timeperiod, after the period is complete
STATE_MACHINE_RECOMPUTING = 'recomputing'

# Job is run multiple times per timeperiod
STATE_MACHINE_CONTINUOUS = 'continuous'

# Job is run multiple times per timeperiod; assumes the
# underlying data is streaming and recomputes the processing boundaries
STATE_MACHINE_DISCRETE = 'discrete'

STATE_MACHINE_FREERUN = 'freerun'

# allows processing of the timeperiod, however finalization of the timeperiod
# is not allowed until all blocking timeperiods are processed
BLOCKING_NORMAL = 'blocking_normal'

# any processing of the timeperiod is blocked until all blocking timeperiods are processed
BLOCKING_DEPENDENCIES = 'blocking_dependencies'

# any processing of the timeperiod is blocked until all children timeperiods are processed
BLOCKING_CHILDREN = 'blocking_children'

# MQ exchanges that are required by Synergy Scheduler
EXCHANGE_MANAGED_WORKER = 'ex_managed_worker'
EXCHANGE_FREERUN_WORKER = 'ex_freerun_worker'
EXCHANGE_UTILS = 'ex_utils'

# DB Collection/Table names
COLLECTION_MANAGED_PROCESS = 'managed_process'
COLLECTION_FREERUN_PROCESS = 'freerun_process'
COLLECTION_UNIT_OF_WORK = 'unit_of_work'
COLLECTION_LOG_RECORDING = 'log_recording'

COLLECTION_JOB_HOURLY = 'job_hourly'
COLLECTION_JOB_DAILY = 'job_daily'
COLLECTION_JOB_MONTHLY = 'job_monthly'
COLLECTION_JOB_YEARLY = 'job_yearly'

QUEUE_UOW_STATUS = 'q_uow_status'
QUEUE_JOB_STATUS = 'q_job_status'



__author__ = 'Bohdan Mushkevych'

from synergy.db.model import job
from synergy.system import time_helper
from synergy.system.immutable_dict import ImmutableDict
from synergy.conf import context


class NodesCompositeState(object):
    """ Instance of this structure represents composite state of TreeNodes """

    def __init__(self):
        super(NodesCompositeState, self).__init__()

        # True if all dependent_on Jobs are finished
        self.all_finished = True

        # True if all dependent_on Jobs are in [STATE_PROCESSED, STATE_NOOP]
        self.all_processed = True

        # True if all dependent_on Jobs are either active or in [STATE_PROCESSED, STATE_NOOP]
        self.all_healthy = True

        # True if among dependent_on periods are some in STATE_SKIPPED
        self.skipped_present = False

    def enlist(self, tree_node):
        assert isinstance(tree_node, TreeNode)
        if not tree_node.job_record.is_finished:
            self.all_finished = False
        if not (tree_node.job_record.is_processed or tree_node.job_record.is_noop):
            self.all_processed = False
        if not (tree_node.job_record.is_active or tree_node.job_record.is_processed or tree_node.job_record.is_noop):
            self.all_healthy = False
        if tree_node.job_record.is_skipped:
            self.skipped_present = True


class AbstractTreeNode(object):
    def __init__(self, tree, parent, process_name, timeperiod, job_record):
        self.tree = tree
        self.parent = parent
        self.process_name = process_name
        self.timeperiod = timeperiod
        self.job_record = job_record

        # fields self.time_qualifier and self.children are properly set in the child class
        self.time_qualifier = None
        self.children = ImmutableDict({})

    def is_finalizable(self):
        """method checks whether:
         - all counterpart of this node in dependent_on trees are finished
         - all direct children of the node are finished
         - the node itself is in active state"""

        composite_state = self.dependent_on_composite_state()
        if not composite_state.all_finished:
            return False

        if self.job_record is None:
            self.tree.timetable.assign_job_record(self)

        children_processed = all([child.job_record.is_finished for child in self.children.values()])

        return children_processed and self.job_record.is_active

    def validate(self):
        """method traverse tree and performs following activities:
        * requests a job record in STATE_EMBRYO if no job record is currently assigned to the node
        * requests nodes for reprocessing, if STATE_PROCESSED node relies on unfinalized nodes
        * requests node for skipping if it is daily node and all 24 of its Hourly nodes are in STATE_SKIPPED state"""

        # step 1: request Job record if current one is not set
        if self.job_record is None:
            self.tree.timetable.assign_job_record(self)

        # step 2: define if current node has a younger sibling
        next_timeperiod = time_helper.increment_timeperiod(self.time_qualifier, self.timeperiod)
        has_younger_sibling = next_timeperiod in self.parent.children

        # step 3: define if all children are done and if perhaps they all are in STATE_SKIPPED
        all_children_skipped = True
        all_children_finished = True
        for timeperiod, child in self.children.items():
            child.validate()

            if child.job_record.is_active:
                all_children_finished = False
            if not child.job_record.is_skipped:
                all_children_skipped = False

        # step 4: request this node's reprocessing if it is enroute to STATE_PROCESSED
        # while some of its children are still performing processing
        if all_children_finished is False and self.job_record.is_finished:
            self.tree.timetable.reprocess_tree_node(self)

        # step 5: verify if this node should be transferred to STATE_SKIPPED
        # algorithm is following:
        # point a: node must have children
        # point b: existence of a younger sibling means that the tree contains another node of the same level
        # thus - should the tree.build_timeperiod be not None - the children level of this node is fully constructed
        # point c: if all children of this node are in STATE_SKIPPED then we will set this node state to STATE_SKIPPED
        if len(self.children) != 0 \
                and all_children_skipped \
                and self.tree.build_timeperiod is not None \
                and has_younger_sibling is True \
                and not self.job_record.is_skipped:
            self.tree.timetable.skip_tree_node(self)

    def add_log_entry(self, entry):
        """ :db.model.job record holds event log, that can be accessed by MX
            this method adds a record and removes oldest one if necessary """
        event_log = self.job_record.event_log
        if len(event_log) > job.MAX_NUMBER_OF_EVENTS:
            del event_log[-1]
        event_log.insert(0, entry)

    def find_counterpart_in(self, tree_b):
        """ Finds a TreeNode counterpart for this node in tree_b
            :param tree_b: target tree that hosts counterpart to this node
            :return: TreeNode from tree_b that has the same timeperiod as self.timeperiod,
                or None if no counterpart ware found
        """

        tree_b_hierarchy_entry = tree_b.process_hierarchy.get_by_qualifier(self.time_qualifier)
        if not tree_b_hierarchy_entry:
            # special case when tree with more levels depends on the tree with smaller amount of levels
            # for example ThreeLevel Financial tree depends on TwoLevel Google Channel
            # in this case - we just verify time-periods that matches in both trees;
            # for levels that have no match, we assume that dependency does not exists
            # for example Financial Monthly has no counterpart in Google Daily Report -
            # so we assume that its not blocked
            node_b = None
        else:
            node_b = tree_b.get_node(tree_b_hierarchy_entry.process_entry.process_name, self.timeperiod)

        return node_b

    def dependent_on_composite_state(self):
        """ method iterates over all nodes that provide dependency to the current node,
            and compile composite state of them all
            :return instance of <NodesCompositeState>
        """
        composite_state = NodesCompositeState()

        for dependent_on in self.tree.dependent_on:
            node_b = self.find_counterpart_in(dependent_on)

            if node_b is None:
                # special case when counterpart tree has no process with corresponding time_qualifier
                # for example Financial Monthly has no counterpart in Third-party Daily Report -
                # so we assume that its not blocked
                continue

            composite_state.enlist(node_b)

        return composite_state


class TreeNode(AbstractTreeNode):
    def __init__(self, tree, parent, process_name, timeperiod, job_record):
        super(TreeNode, self).__init__(tree, parent, process_name, timeperiod, job_record)
        self.time_qualifier = context.process_context[process_name].time_qualifier

        child_hierarchy_entry = tree.process_hierarchy.get_child_by_qualifier(self.time_qualifier)
        if child_hierarchy_entry:
            children = dict()
        else:
            # this is the bottom process of the process hierarchy with no children
            children = ImmutableDict({})
        self.children = children


class RootNode(AbstractTreeNode):
    def __init__(self, tree):
        super(RootNode, self).__init__(tree, None, None, None, None)
        self.time_qualifier = None
        self.children = dict()



__author__ = 'Bohdan Mushkevych'

from six import string_types
from odm.document import BaseDocument
from odm.fields import StringField, ListField

TREE_NAME = 'tree_name'
DEPENDENT_ON = 'dependent_on'
ENCLOSED_PROCESSES = 'enclosed_processes'
MX_PAGE = 'mx_page'
MX_NAME = 'mx_name'


class TimetableTreeEntry(BaseDocument):
    """ Non-persistent model. Class presents single process tree (an atomic entry for the Timetable) """

    tree_name = StringField(TREE_NAME)
    dependent_on = ListField(DEPENDENT_ON)
    enclosed_processes = ListField(ENCLOSED_PROCESSES)
    mx_name = StringField(MX_NAME)
    mx_page = StringField(MX_PAGE)

    @BaseDocument.key.getter
    def key(self):
        return self.tree_name

    @key.setter
    def key(self, value):
        self.tree_name = value


def timetable_tree_entry(tree_name,
                         enclosed_processes,
                         dependent_on=None,
                         mx_name=None,
                         mx_page=None):
    """ creates timetable context entry """
    assert enclosed_processes is not None and not isinstance(enclosed_processes, string_types)
    assert dependent_on is not None and not isinstance(dependent_on, string_types)

    timetable_entry = TimetableTreeEntry(tree_name=tree_name,
                                         enclosed_processes=enclosed_processes,
                                         dependent_on=dependent_on,
                                         mx_name=mx_name,
                                         mx_page=mx_page)
    return timetable_entry



__author__ = 'Bohdan Mushkevych'

from synergy.mx.base_request_handler import BaseRequestHandler, valid_action_request


class GcActionHandler(BaseRequestHandler):
    def __init__(self, request, **values):
        super(GcActionHandler, self).__init__(request, **values)
        self.process_name = self.request_arguments.get('process_name')
        self.is_request_valid = True if self.process_name else False

        if self.is_request_valid:
            self.process_name = self.process_name.strip()

    def action_refresh(self):
        self.scheduler.gc.validate()
        self.scheduler.gc.scan_uow_candidates()
        self.logger.info('MX: performed GC Refresh')
        return self.reply_ok()

    def action_flush_all(self):
        self.scheduler.gc.flush(ignore_priority=True)
        self.logger.info('MX: performed GC Flush All')
        return self.reply_ok()

    @valid_action_request
    def action_flush_one(self):
        self.scheduler.gc.flush_one(process_name=self.process_name, ignore_priority=True)
        self.logger.info('MX: performed GC Flush for {0}'.format(self.process_name))
        return self.reply_ok()



__author__ = 'Bohdan Mushkevych'

from synergy.system.event_clock import EventClock
from synergy.system.repeat_timer import RepeatTimer

TRIGGER_PREAMBLE_AT = 'at '
TRIGGER_PREAMBLE_EVERY = 'every '


def parse_time_trigger_string(trigger_frequency):
    """
    :param trigger_frequency: human-readable and editable string in one of two formats:
     - 'at Day_of_Week-HH:MM, ..., Day_of_Week-HH:MM'
     - 'every NNN'
    :return: return tuple (parsed_trigger_frequency, timer_klass)
    """
    # replace multiple spaces with one
    trigger_frequency = ' '.join(trigger_frequency.split())

    if trigger_frequency.startswith(TRIGGER_PREAMBLE_AT):
        # EventClock block
        trigger_frequency = trigger_frequency[len(TRIGGER_PREAMBLE_AT):]
        parsed_trigger_frequency = trigger_frequency.replace(' ', '').replace(',', ' ').split(' ')
        timer_klass = EventClock
    elif trigger_frequency.startswith(TRIGGER_PREAMBLE_EVERY):
        # RepeatTimer block
        trigger_frequency = trigger_frequency[len(TRIGGER_PREAMBLE_EVERY):]
        parsed_trigger_frequency = int(trigger_frequency)
        timer_klass = RepeatTimer
    else:
        raise ValueError('Unknown time trigger format {0}'.format(trigger_frequency))

    return parsed_trigger_frequency, timer_klass


def format_time_trigger_string(timer_instance):
    """
    :param timer_instance: either instance of RepeatTimer or EventClock
    :return: human-readable and editable string in one of two formats:
     - 'at Day_of_Week-HH:MM, ..., Day_of_Week-HH:MM'
     - 'every NNN'
    """
    if isinstance(timer_instance, RepeatTimer):
        return TRIGGER_PREAMBLE_EVERY + str(timer_instance.interval_new)
    elif isinstance(timer_instance, EventClock):
        timestamps = [repr(x) for x in timer_instance.timestamps]
        return TRIGGER_PREAMBLE_AT + ','.join(timestamps)
    else:
        raise ValueError('Unknown timer instance type {0}'.format(timer_instance.__class__.__name__))



__author__ = 'Bohdan Mushkevych'

from six import string_types
from odm.document import BaseDocument
from odm.fields import StringField, ObjectIdField

PROCESS_NAME = 'process_name'           # name of the process to handle the schedulables
ENTRY_NAME = 'entry_name'               # name of the schedulable, if applicable
RECORD_DB_ID = 'record_db_id'           # associated with either UOW.db_id or Job.db_id


class MqTransmission(BaseDocument):
    """ Non-persistent model. Instance of this class presents either:
     - single request from Synergy Scheduler to any worker
     - response/report from the worker to the Synergy Scheduler """

    process_name = StringField(PROCESS_NAME)
    entry_name = StringField(ENTRY_NAME, null=True)
    record_db_id = ObjectIdField(RECORD_DB_ID)

    @BaseDocument.key.getter
    def key(self):
        return self.process_name, self.entry_name

    @key.setter
    def key(self, value):
        if not isinstance(value, string_types):
            self.process_name = value[0]
            self.entry_name = value[1]
        else:
            self.process_name = value
            self.entry_name = None

    def __str__(self):
        return '{0}::{1}#{2}'.format(self.process_name, self.entry_name, self.record_db_id)



settings = dict(
    mq_timeout_sec=300.0,
    mq_durable=True,
    mq_exclusive=False,
    mq_auto_delete=False,
    mq_delivery_mode=2,
    mq_no_ack=False,

    gc_run_interval=60,          # number of seconds between GarbageCollector runs
    gc_life_support_hours=48,    # number of hours from UOW creation time to keep UOW re-posting to MQ
    gc_resubmit_after_hours=1,   # number of hours, GC waits for the worker to pick up the UOW from MQ before re-posting
    gc_release_lag_minutes=15,   # number of minutes, GC keeps the UOW in the queue before posting it into MQ

    mx_host='0.0.0.0',           # management extension host (0.0.0.0 opens all interfaces)
    mx_port=5000,                # management extension port
    mx_children_limit=168,       # maximum number of children at any given level returned by MX
    perf_ticker_interval=60,     # seconds between performance ticker messages
)



__author__ = 'Bohdan Mushkevych'


class DuplicateKeyError(Exception):
    def __init__(self, process_name, timeperiod, start_id, end_id, *args, **kwargs):
        self.process_name = process_name
        self.timeperiod = timeperiod
        self.start_id = start_id
        self.end_id = end_id
        super(DuplicateKeyError, self).__init__(*args, **kwargs)



from __future__ import print_function

__author__ = 'Bohdan Mushkevych'

import os
import sys
import gzip
import hashlib

from six import string_types

from synergy.conf import settings
from synergy.conf import context


def fully_qualified_table_name(table_name):
    # fully qualified table name
    fqtn = settings.settings['aws_redshift_orca_prefix'] + table_name + settings.settings['aws_redshift_orca_suffix']
    return fqtn


def create_s3_file_uri(s3_bucket, timeperiod, file_name):
    return 's3://{0}/{1}/{2}'.format(s3_bucket, timeperiod, file_name)


def break_s3_file_uri(fully_qualified_file):
    """
    :param fully_qualified_file: in form s3://{0}/{1}/{2}
    where {0} is bucket name, {1} is timeperiod and {2} - file name
    :return: tuple (s3://{0}, {1}/{2})
    """
    tokens = fully_qualified_file.rsplit('/', 2)
    return tokens[0], '{0}/{1}'.format(tokens[1], tokens[2])


def unicode_truncate(s, length, encoding='utf-8'):
    encoded = s.encode(encoding)[:length]
    return encoded.decode(encoding, errors='ignore')


def compute_gzip_md5(file_name):
    """ method traverses compressed file and calculates its MD5 checksum """
    md5 = hashlib.md5()
    file_obj = gzip.open(file_name, 'rb')
    for chunk in iter(lambda: file_obj.read(8192), ''):
        md5.update(chunk)

    file_obj.close()
    return md5.hexdigest()


def increment_family_property(key, family):
    if key is None:
        return

    if not isinstance(key, string_types):
        key = str(key)

    if key in family:
        family[key] += 1
    else:
        family[key] = 1


def copy_and_sum_families(family_source, family_target):
    """ methods iterates thru source family and copies its entries to target family
    in case key already exists in both families - then the values are added"""
    for every in family_source:
        if every not in family_target:
            family_target[every] = family_source[every]
        else:
            family_target[every] += family_source[every]


def get_pid_filename(process_name):
    """method returns path for the PID FILENAME """
    return settings.settings['pid_directory'] + context.process_context[process_name].pid_filename


def create_pid_file(process_name):
    """ creates pid file and writes os.pid() in there """
    pid_filename = get_pid_filename(process_name)
    try:
        with open(pid_filename, mode='w') as pid_file:
            pid_file.write(str(os.getpid()))
    except Exception as e:
        print('Unable to create pid file at: {0}, because of: {1}'.format(pid_filename, e),
              file=sys.stderr)


def remove_pid_file(process_name):
    """ removes pid file """
    pid_filename = get_pid_filename(process_name)
    try:
        os.remove(pid_filename)
        print('Removed pid file at: {0}'.format(pid_filename), file=sys.stdout)
    except Exception as e:
        print('Unable to remove pid file at: {0}, because of: {1}'.format(pid_filename, e),
              file=sys.stderr)



__author__ = 'Bohdan Mushkevych'

from odm.fields import StringField, ObjectIdField, BooleanField, IntegerField

from synergy.db.model.daemon_process_entry import DaemonProcessEntry
from synergy.scheduler.scheduler_constants import BLOCKING_CHILDREN, BLOCKING_DEPENDENCIES, BLOCKING_NORMAL, \
    EXCHANGE_MANAGED_WORKER, STATE_MACHINE_DISCRETE


PROCESS_NAME = 'process_name'
IS_ON = 'is_on'
TRIGGER_FREQUENCY = 'trigger_frequency'
STATE_MACHINE_NAME = 'state_machine_name'
BLOCKING_TYPE = 'blocking_type'
SOURCE = 'source'
SINK = 'sink'
TIME_QUALIFIER = 'time_qualifier'
TIME_GROUPING = 'time_grouping'


class ManagedProcessEntry(DaemonProcessEntry):
    """ Class presents single configuration entry for scheduler managed (i.e. - non-freerun) processes """
    db_id = ObjectIdField('_id', null=True)
    source = StringField(SOURCE, null=True)
    sink = StringField(SINK, null=True)
    time_qualifier = StringField(TIME_QUALIFIER)
    time_grouping = IntegerField(TIME_GROUPING)
    trigger_frequency = StringField(TRIGGER_FREQUENCY)
    is_on = BooleanField(IS_ON, default=False)
    state_machine_name = StringField(STATE_MACHINE_NAME)
    blocking_type = StringField(BLOCKING_TYPE, choices=[BLOCKING_CHILDREN, BLOCKING_DEPENDENCIES, BLOCKING_NORMAL])

    @DaemonProcessEntry.key.getter
    def key(self):
        return self.process_name

    @DaemonProcessEntry.key.setter
    def key(self, value):
        self.process_name = value


def managed_context_entry(process_name,
                          classname,
                          token,
                          time_qualifier,
                          trigger_frequency='every 60',
                          state_machine_name=STATE_MACHINE_DISCRETE,
                          is_on=True,
                          exchange=EXCHANGE_MANAGED_WORKER,
                          blocking_type=BLOCKING_NORMAL,
                          present_on_boxes=None,
                          time_grouping=1,
                          arguments=None,
                          queue=None,
                          routing=None,
                          source=None,
                          sink=None,
                          pid_file=None,
                          log_file=None):
    """ forms process context entry """
    _ROUTING_PREFIX = 'routing_'
    _QUEUE_PREFIX = 'queue_'

    if arguments is not None:
        assert isinstance(arguments, dict)

    process_entry = ManagedProcessEntry(
        process_name=process_name,
        trigger_frequency=trigger_frequency,
        state_machine_name=state_machine_name,
        is_on=is_on,
        blocking_type=blocking_type,
        classname=classname,
        token=token,
        source=source,
        sink=sink,
        mq_queue=queue if queue is not None else _QUEUE_PREFIX + token + time_qualifier,
        mq_routing_key=routing if routing is not None else _ROUTING_PREFIX + token + time_qualifier,
        mq_exchange=exchange,
        present_on_boxes=present_on_boxes,
        arguments=arguments if arguments is not None else dict(),
        time_qualifier=time_qualifier,
        time_grouping=time_grouping,
        log_filename=log_file if log_file is not None else token + time_qualifier + '.log',
        pid_filename=pid_file if pid_file is not None else token + time_qualifier + '.pid')
    return process_entry



__author__ = 'Bohdan Mushkevych'

from threading import Thread

from amqp import AMQPError

from synergy.db.model.mq_transmission import MqTransmission
from synergy.db.dao.job_dao import JobDao
from synergy.scheduler.scheduler_constants import QUEUE_JOB_STATUS
from synergy.scheduler.thread_handler import ManagedThreadHandler
from synergy.mq.flopsy import Consumer


class JobStatusListener(object):
    """ class instance listens to the QUEUE_JOB_STATUS queue and triggers ManagedThreadHandlers if applicable """

    def __init__(self, scheduler):
        self.scheduler = scheduler
        self.timetable = scheduler.timetable
        self.logger = scheduler.logger
        self.job_dao = JobDao(self.logger)
        self.consumer = Consumer(QUEUE_JOB_STATUS)
        self.main_thread = None

    def __del__(self):
        try:
            self.logger.info('JobStatusListener: Closing Flopsy Consumer...')
            self.consumer.close()
        except Exception as e:
            self.logger.error('JobStatusListener: Exception caught while closing Flopsy Consumer: {0}'.format(e))

    # ********************** thread-related methods ****************************
    def _mq_callback(self, message):
        """ method receives a message from Synergy Scheduler notifying of Job completion,
            builds up a list of dependant TreeNodes/Jobs and triggers their ManagedThreadHandlers, if applicable
            :param message: <MqTransmission> mq message """
        try:
            self.logger.info('JobStatusListener {')

            mq_request = MqTransmission.from_json(message.body)
            job_record = self.job_dao.get_by_id(mq_request.process_name, mq_request.record_db_id)

            # step 1: identify dependant tree nodes
            tree_obj = self.timetable.get_tree(job_record.process_name)
            tree_node = tree_obj.get_node(job_record.process_name, job_record.timeperiod)
            dependant_nodes = self.timetable._find_dependant_tree_nodes(tree_node)

            # step 2: form list of handlers to trigger
            handlers_to_trigger = set()
            for node in dependant_nodes:
                state_machine = self.scheduler.state_machine_for(node.process_name)
                if state_machine.run_on_active_timeperiod:
                    # ignore dependant processes whose state machine can run on an active timeperiod
                    continue
                handlers_to_trigger.add(self.scheduler.managed_handlers[node.process_name])

            # step 3: iterate the list of handlers and trigger them
            for handler in handlers_to_trigger:
                assert isinstance(handler, ManagedThreadHandler)
                handler.trigger()

        except KeyError:
            self.logger.error('Access error for {0}'.format(message.body), exc_info=True)
        except Exception:
            self.logger.error('Error during ManagedThreadHandler.trigger call {0}'.format(message.body), exc_info=True)
        finally:
            self.consumer.acknowledge(message.delivery_tag)
            self.logger.info('JobStatusListener }')

    def _run_mq_listener(self):
        try:
            self.consumer.register(self._mq_callback)
            self.logger.info('JobStatusListener: instantiated and activated.')
            self.consumer.wait()
        except (AMQPError, IOError) as e:
            self.logger.error('JobStatusListener: AMQPError {0}'.format(e))
        finally:
            self.__del__()
            self.logger.info('JobStatusListener: Shut down.')

    def start(self, *_):
        self.main_thread = Thread(target=self._run_mq_listener)
        self.main_thread.daemon = True
        self.main_thread.start()

    def stop(self):
        """ method stops currently MQ Consumer listener, if any """
        self.__del__()



__author__ = 'Bohdan Mushkevych'

from odm.document import BaseDocument
from odm.fields import StringField

MQ_QUEUE = 'mq_queue'
MQ_EXCHANGE = 'mq_exchange'
MQ_ROUTING_KEY = 'mq_routing_key'


class QueueContextEntry(BaseDocument):
    """ Non-persistent model. This class presents Queue Context Entry record """

    mq_queue = StringField(MQ_QUEUE)
    mq_exchange = StringField(MQ_EXCHANGE)
    mq_routing_key = StringField(MQ_ROUTING_KEY)

    @BaseDocument.key.getter
    def key(self):
        return self.mq_queue

    @key.setter
    def key(self, value):
        """ :param value: name of the mq queue """
        self.mq_queue = value


def queue_context_entry(exchange,
                        queue_name,
                        routing=None):
    """ forms queue's context entry """
    if routing is None:
        routing = queue_name

    queue_entry = QueueContextEntry(mq_queue=queue_name,
                                    mq_exchange=exchange,
                                    mq_routing_key=routing)
    return queue_entry



__author__ = 'Bohdan Mushkevych'

from synergy.db.dao.base_dao import BaseDao
from synergy.db.model.managed_process_entry import ManagedProcessEntry, PROCESS_NAME
from synergy.system.decorator import thread_safe
from synergy.scheduler.scheduler_constants import COLLECTION_MANAGED_PROCESS


class ManagedProcessDao(BaseDao):
    """ Thread-safe Data Access Object for managed_process table/collection """

    def __init__(self, logger):
        super(ManagedProcessDao, self).__init__(logger=logger,
                                                model_class=ManagedProcessEntry,
                                                primary_key=[PROCESS_NAME],
                                                collection_name=COLLECTION_MANAGED_PROCESS)

    @thread_safe
    def clear(self):
        """ removes all documents in this collection """
        collection = self.ds.connection(COLLECTION_MANAGED_PROCESS)
        return collection.delete_many(filter={})



__author__ = 'Bohdan Mushkevych'

import collections

from synergy.system import time_helper
from synergy.conf import context
from synergy.mx.base_request_handler import valid_action_request
from synergy.mx.abstract_action_handler import AbstractActionHandler
from synergy.mx.tree_node_details import TreeNodeDetails


class ManagedActionHandler(AbstractActionHandler):
    def __init__(self, request, **values):
        super(ManagedActionHandler, self).__init__(request, **values)
        self.process_name = self.request_arguments.get('process_name')
        self.timeperiod = self.request_arguments.get('timeperiod')
        self.is_request_valid = True if self.process_name and self.timeperiod else False

        if self.is_request_valid:
            self.process_name = self.process_name.strip()
            self.timeperiod = self.timeperiod.strip()

    def _get_tree_node(self):
        tree = self.scheduler.timetable.get_tree(self.process_name)
        if tree is None:
            raise UserWarning('No Timetable tree is registered for process {0}'.format(self.process_name))

        time_qualifier = context.process_context[self.process_name].time_qualifier
        self.timeperiod = time_helper.cast_to_time_qualifier(time_qualifier, self.timeperiod)
        node = tree.get_node(self.process_name, self.timeperiod)
        return node

    @AbstractActionHandler.thread_handler.getter
    def thread_handler(self):
        handler_key = self.process_name
        return self.scheduler.managed_handlers[handler_key]

    @AbstractActionHandler.process_entry.getter
    def process_entry(self):
        return self.thread_handler.process_entry

    @AbstractActionHandler.uow_id.getter
    def uow_id(self):
        node = self._get_tree_node()
        return None if not node.job_record else node.job_record.related_unit_of_work

    @valid_action_request
    def action_reprocess(self):
        node = self._get_tree_node()

        msg = 'MX: requesting REPROCESS for {0} in timeperiod {1}'.format(self.process_name, self.timeperiod)
        self.scheduler.timetable.add_log_entry(self.process_name, self.timeperiod, msg)
        self.logger.info(msg + ' {')

        tx_context = self.scheduler.timetable.reprocess_tree_node(node)
        self.scheduler.gc.validate()

        resp = collections.defaultdict(dict)
        for process_name, nodes_context in tx_context.items():
            for timeperiod, node in nodes_context.items():
                resp[process_name][timeperiod] = TreeNodeDetails.get_details(node)
        self.logger.info('MX }')
        return resp

    @valid_action_request
    def action_skip(self):
        node = self._get_tree_node()

        msg = 'MX: requesting SKIP for {0} in timeperiod {1}'.format(self.process_name, self.timeperiod)
        self.scheduler.timetable.add_log_entry(self.process_name, self.timeperiod, msg)
        self.logger.info(msg + ' {')

        tx_context = self.scheduler.timetable.skip_tree_node(node)
        self.scheduler.gc.validate()

        resp = collections.defaultdict(dict)
        for process_name, nodes_context in tx_context.items():
            for timeperiod, node in nodes_context.items():
                resp[process_name][timeperiod] = TreeNodeDetails.get_details(node)
        self.logger.info('MX }')
        return resp

    @valid_action_request
    def action_get_event_log(self):
        node = self._get_tree_node()
        return {'event_log': [] if not node.job_record else node.job_record.event_log}



__author__ = 'Bohdan Mushkevych'

import collections
from datetime import datetime
from threading import RLock

from synergy.db.dao.job_dao import JobDao
from synergy.db.model.job import Job
from synergy.conf import context
from synergy.conf import settings
from synergy.system import time_helper, utils
from synergy.system.time_qualifier import *
from synergy.system.decorator import thread_safe
from synergy.scheduler.scheduler_constants import COLLECTION_JOB_HOURLY, COLLECTION_JOB_DAILY, \
    COLLECTION_JOB_MONTHLY, COLLECTION_JOB_YEARLY
from synergy.scheduler.tree import MultiLevelTree
from synergy.scheduler.state_machine_recomputing import StateMachineRecomputing
from synergy.scheduler.state_machine_continuous import StateMachineContinuous
from synergy.scheduler.state_machine_discrete import StateMachineDiscrete
from synergy.scheduler.state_machine_freerun import StateMachineFreerun


class Timetable(object):
    """ Timetable holds all known process trees, where every node presents a timeperiod-driven job"""

    def __init__(self, logger):
        self.lock = RLock()
        self.logger = logger
        self.job_dao = JobDao(self.logger)

        # state_machines must be constructed before the trees
        self.state_machines = self._construct_state_machines()

        # self.trees contain all of the trees and manages much of their life cycle
        # remember to enlist here all trees the system is working with
        self.trees = self._construct_trees_from_context()

        self._register_dependencies()
        self.load_tree()
        self.build_trees()
        self.validate()

    def _construct_state_machines(self):
        """ :return: dict in format <state_machine_common_name: instance_of_the_state_machine> """
        state_machines = dict()
        for state_machine in [StateMachineRecomputing(self.logger, self),
                              StateMachineContinuous(self.logger, self),
                              StateMachineDiscrete(self.logger, self),
                              StateMachineFreerun(self.logger)]:
            state_machines[state_machine.name] = state_machine
        return state_machines

    def _construct_trees_from_context(self):
        trees = dict()
        for tree_name, context_entry in context.timetable_context.items():
            tree = MultiLevelTree(process_names=context_entry.enclosed_processes,
                                  timetable=self,
                                  tree_name=tree_name,
                                  mx_name=context_entry.mx_name,
                                  mx_page=context_entry.mx_page)
            trees[tree_name] = tree
        return trees

    def _register_dependencies(self):
        """ register dependencies between trees"""
        for tree_name, context_entry in context.timetable_context.items():
            tree = self.trees[tree_name]
            assert isinstance(tree, MultiLevelTree)
            for dependent_on in context_entry.dependent_on:
                dependent_on_tree = self.trees[dependent_on]
                assert isinstance(dependent_on_tree, MultiLevelTree)
                tree.register_dependent_on(dependent_on_tree)

    # *** node manipulation methods ***
    def _find_dependant_trees(self, tree_obj):
        """ returns list of trees that are dependent_on given tree_obj """
        dependant_trees = []
        for tree_name, tree in self.trees.items():
            if tree_obj in tree.dependent_on:
                dependant_trees.append(tree)
        return dependant_trees

    def _find_dependant_tree_nodes(self, node_a):
        dependant_nodes = set()
        for tree_b in self._find_dependant_trees(node_a.tree):
            node_b = node_a.find_counterpart_in(tree_b)
            if node_b is None:
                continue
            dependant_nodes.add(node_b)
        return dependant_nodes

    @thread_safe
    def reprocess_tree_node(self, tree_node, tx_context=None):
        """ method reprocesses the node and all its dependants and parent nodes """
        if not tx_context:
            # create transaction context if one was not provided
            # format: {process_name: {timeperiod: AbstractTreeNode} }
            tx_context = collections.defaultdict(dict)

        if tree_node.parent is None:
            # do not process 'root' - the only node that has None as 'parent'
            return tx_context
        if tree_node.timeperiod in tx_context[tree_node.process_name]:
            # the node has already been marked for re-processing
            return tx_context

        if tree_node.job_record.is_embryo:
            # the node does not require re-processing
            pass
        else:
            state_machine_name = context.process_context[tree_node.process_name].state_machine_name
            state_machine = self.state_machines[state_machine_name]
            state_machine.reprocess_job(tree_node.job_record)

        tx_context[tree_node.process_name][tree_node.timeperiod] = tree_node
        self.reprocess_tree_node(tree_node.parent, tx_context)

        dependant_nodes = self._find_dependant_tree_nodes(tree_node)
        for node in dependant_nodes:
            self.reprocess_tree_node(node, tx_context)

        return tx_context

    @thread_safe
    def skip_tree_node(self, tree_node, tx_context=None):
        """ method skips the node and all its dependants and child nodes """
        if not tx_context:
            # create transaction context if one was not provided
            # format: {process_name: {timeperiod: AbstractTreeNode} }
            tx_context = collections.defaultdict(dict)

        if tree_node.timeperiod in tx_context[tree_node.process_name]:
            # the node has already been marked for skipping
            return tx_context

        if tree_node.job_record.is_finished:
            # the node is finished and does not require skipping
            pass
        else:
            state_machine_name = context.process_context[tree_node.process_name].state_machine_name
            state_machine = self.state_machines[state_machine_name]
            state_machine.skip_job(tree_node.job_record)

        tx_context[tree_node.process_name][tree_node.timeperiod] = tree_node
        for timeperiod, node in tree_node.children.items():
            self.skip_tree_node(node, tx_context)

        dependant_nodes = self._find_dependant_tree_nodes(tree_node)
        for node in dependant_nodes:
            self.skip_tree_node(node, tx_context)

        return tx_context

    @thread_safe
    def assign_job_record(self, tree_node):
        """ - looks for an existing job record in the DB, and if not found
            - creates a job record in STATE_EMBRYO and bind it to the given tree node """
        try:
            job_record = self.job_dao.get_one(tree_node.process_name, tree_node.timeperiod)
        except LookupError:
            state_machine_name = context.process_context[tree_node.process_name].state_machine_name
            state_machine = self.state_machines[state_machine_name]
            job_record = state_machine.create_job(tree_node.process_name, tree_node.timeperiod)
        tree_node.job_record = job_record

    # *** Tree-manipulation methods ***
    @thread_safe
    def get_tree(self, process_name):
        """ return tree that is managing time-periods for given process"""
        for tree_name, tree in self.trees.items():
            if process_name in tree:
                return tree

    @thread_safe
    def _build_tree_by_level(self, time_qualifier, collection_name, since):
        """ method iterated thru all documents in all job collections and builds a tree of known system state"""
        invalid_tree_records = dict()
        invalid_tq_records = dict()

        try:
            job_records = self.job_dao.get_all(collection_name, since)
            for job_record in job_records:
                tree = self.get_tree(job_record.process_name)
                if tree is None:
                    utils.increment_family_property(job_record.process_name, invalid_tree_records)
                    continue

                job_time_qualifier = context.process_context[job_record.process_name].time_qualifier
                if time_qualifier != job_time_qualifier:
                    utils.increment_family_property(job_record.process_name, invalid_tq_records)
                    continue

                tree.update_node(job_record)

        except LookupError:
            self.logger.warning('No job records in {0}.'.format(collection_name))

        for name, counter in invalid_tree_records.items():
            self.logger.warning('Skipping {0} job records for {1} since no tree is handling it.'
                                .format(counter, name))

        for name, counter in invalid_tq_records.items():
            self.logger.warning('Skipping {0} job records for {1} since the process has different time qualifier.'
                                .format(counter, name))

    @thread_safe
    def load_tree(self):
        """ method iterates thru all objects older than synergy_start_timeperiod parameter in job collections
        and loads them into this timetable"""
        timeperiod = settings.settings['synergy_start_timeperiod']
        yearly_timeperiod = time_helper.cast_to_time_qualifier(QUALIFIER_YEARLY, timeperiod)
        monthly_timeperiod = time_helper.cast_to_time_qualifier(QUALIFIER_MONTHLY, timeperiod)
        daily_timeperiod = time_helper.cast_to_time_qualifier(QUALIFIER_DAILY, timeperiod)
        hourly_timeperiod = time_helper.cast_to_time_qualifier(QUALIFIER_HOURLY, timeperiod)

        self._build_tree_by_level(QUALIFIER_HOURLY, COLLECTION_JOB_HOURLY, since=hourly_timeperiod)
        self._build_tree_by_level(QUALIFIER_DAILY, COLLECTION_JOB_DAILY, since=daily_timeperiod)
        self._build_tree_by_level(QUALIFIER_MONTHLY, COLLECTION_JOB_MONTHLY, since=monthly_timeperiod)
        self._build_tree_by_level(QUALIFIER_YEARLY, COLLECTION_JOB_YEARLY, since=yearly_timeperiod)

    @thread_safe
    def build_trees(self):
        """ method iterates thru all trees and ensures that all time-period nodes are created up till <utc_now>"""
        for tree_name, tree in self.trees.items():
            tree.build_tree()

    @thread_safe
    def validate(self):
        """validates that none of nodes in tree is improperly finalized and that every node has job_record"""
        for tree_name, tree in self.trees.items():
            tree.validate()

    @thread_safe
    def dependent_on_composite_state(self, job_record):
        """ :return instance of <NodesCompositeState> """
        assert isinstance(job_record, Job)
        tree = self.get_tree(job_record.process_name)
        node = tree.get_node(job_record.process_name, job_record.timeperiod)
        return node.dependent_on_composite_state()

    # *** Job manipulation methods ***
    @thread_safe
    def skip_if_needed(self, job_record):
        """ method is called from abstract_state_machine.manage_job to notify about job's failed processing
            if should_skip_node returns True - the node's job_record is transferred to STATE_SKIPPED """
        tree = self.get_tree(job_record.process_name)
        node = tree.get_node(job_record.process_name, job_record.timeperiod)
        if tree.should_skip_tree_node(node):
            self.skip_tree_node(node)

    @thread_safe
    def get_next_job_record(self, process_name):
        """ :returns: the next job record to work on for the given process"""
        tree = self.get_tree(process_name)
        node = tree.get_next_node(process_name)

        if node.job_record is None:
            self.assign_job_record(node)
        return node.job_record

    @thread_safe
    def is_job_record_finalizable(self, job_record):
        """ :return: True, if the node and all its children are in [STATE_PROCESSED, STATE_SKIPPED, STATE_NOOP] """
        assert isinstance(job_record, Job)
        tree = self.get_tree(job_record.process_name)
        node = tree.get_node(job_record.process_name, job_record.timeperiod)
        return node.is_finalizable()

    @thread_safe
    def add_log_entry(self, process_name, timeperiod, msg):
        """ adds a non-persistent log entry to the tree node """
        tree = self.get_tree(process_name)
        node = tree.get_node(process_name, timeperiod)
        node.add_log_entry([datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S'), msg])



__author__ = 'Bohdan Mushkevych'

import os
import time
import threading
import fabric.operations
from datetime import datetime

from synergy.conf import settings
from synergy.workers.abstract_mq_worker import AbstractMqWorker
from synergy.workers.worker_constants import *
from synergy.db.model import unit_of_work
from synergy.db.model.mq_transmission import MqTransmission
from synergy.db.dao.unit_of_work_dao import UnitOfWorkDao


class BashRunnable(threading.Thread):
    """Process starts remote or local bash script job, supervises its execution and updates mq"""

    def __init__(self, logger, message, consumer, performance_tracker):
        self.logger = logger
        self.message = message
        self.mq_request = MqTransmission.from_json(message.body)
        self.consumer = consumer
        self.performance_tracker = performance_tracker
        self.alive = False
        self.return_code = -1
        self.uow_dao = UnitOfWorkDao(self.logger)

        self.thread_name = str(self.mq_request)
        super(BashRunnable, self).__init__(name=self.thread_name)

    def _poll_process(self):
        return self.alive, self.return_code

    def _start_process(self):
        try:
            uow = self.uow_dao.get_one(self.mq_request.record_db_id)
            if not uow.is_requested:
                # accept only UOW in STATE_REQUESTED
                self.logger.warning('Skipping UOW: id {0}; state {1};'.format(self.message.body, uow.state),
                                    exc_info=False)
                self.consumer.acknowledge(self.message.delivery_tag)
                return
        except Exception:
            self.logger.error('Safety fuse. Can not identify UOW {0}'.format(self.message.body), exc_info=True)
            self.consumer.acknowledge(self.message.delivery_tag)
            return

        try:
            self.logger.info('start: {0} {{'.format(self.thread_name))
            self.alive = True

            uow.state = unit_of_work.STATE_IN_PROGRESS
            uow.started_at = datetime.utcnow()
            self.uow_dao.update(uow)

            fabric.operations.env.warn_only = True
            fabric.operations.env.abort_on_prompts = True
            fabric.operations.env.use_ssh_config = True
            fabric.operations.env.host_string = uow.arguments[ARGUMENT_CMD_HOST]

            command = os.path.join(uow.arguments[ARGUMENT_CMD_PATH],
                                   uow.arguments[ARGUMENT_CMD_FILE])
            command += ' {0}'.format(uow.arguments[ARGUMENT_CMD_ARGS])

            run_result = fabric.operations.run(command, pty=False)
            if run_result.succeeded:
                self.return_code = 0

            uow.finished_at = datetime.utcnow()
            uow.state = unit_of_work.STATE_PROCESSED
            self.uow_dao.update(uow)

            self.logger.info('Completed {0} with result = {1}'.format(self.thread_name, self.return_code))
        except Exception:
            self.logger.error('Exception on starting: {0}'.format(self.thread_name), exc_info=True)
            uow.state = unit_of_work.STATE_INVALID
            self.uow_dao.update(uow)
        finally:
            self.logger.info('}')
            self.alive = False

    def run(self):
        try:
            self._start_process()
            code = None
            alive = True
            while alive:
                alive, code = self._poll_process()
                time.sleep(0.1)

            if code == 0:
                self.performance_tracker.tracker.increment_success()
            else:
                self.performance_tracker.tracker.increment_failure()

            self.logger.info('BashDriver for {0} return code is {1}'.format(self.thread_name, code))
        except Exception as e:
            self.performance_tracker.tracker.increment_failure()
            self.logger.error('Safety fuse while processing request {0}: {1}'.format(self.message.body, e),
                              exc_info=True)
        finally:
            self.consumer.acknowledge(self.message.delivery_tag)


class BashDriver(AbstractMqWorker):
    """Process facilitates threads running local or remote bash scripts"""

    def __init__(self, process_name):
        super(BashDriver, self).__init__(process_name)
        self.initial_thread_count = threading.active_count()

    def _mq_callback(self, message):
        """ reads JSON request from the mq message and delivers it for processing """
        while threading.active_count() > settings.settings['bash_runnable_count'] + self.initial_thread_count:
            time.sleep(0.1)

        t = BashRunnable(self.logger, message, self.consumer, self.performance_tracker)
        t.daemon = True
        t.start()


if __name__ == '__main__':
    source = BashDriver(PROCESS_BASH_DRIVER)
    source.start()



__author__ = 'Bohdan Mushkevych'

from odm.document import BaseDocument
from odm.fields import StringField, ObjectIdField, IntegerField, BooleanField

BOX_ID = 'box_id'
PROCESS_NAME = 'process_name'
PID = 'pid'
IS_ON = 'is_on'


class BoxConfiguration(BaseDocument):
    """ Persistent model: identifies the machine name and the process where it should be executing """
    db_id = ObjectIdField('_id', null=True)
    box_id = StringField(BOX_ID)
    process_name = StringField(PROCESS_NAME)
    is_on = BooleanField(IS_ON, default=False)
    pid = IntegerField(PID, null=True)



__author__ = 'Bohdan Mushkevych'

from datetime import datetime
from logging import INFO, WARNING, ERROR

from synergy.db.model import job
from synergy.db.model.job import Job
from synergy.db.error import DuplicateKeyError
from synergy.db.dao.unit_of_work_dao import UnitOfWorkDao
from synergy.db.dao.job_dao import JobDao
from synergy.db.model import unit_of_work
from synergy.db.model.unit_of_work import UnitOfWork
from synergy.system.mq_transmitter import MqTransmitter
from synergy.conf import context
from synergy.system.decorator import with_reconnect
from synergy.system import time_helper
from synergy.scheduler.tree_node import NodesCompositeState


class AbstractStateMachine(object):
    """ Abstract state machine used to host common logic for the rest of registered State Machines """

    def __init__(self, logger, timetable, name):
        self.name = name
        self.logger = logger
        self.mq_transmitter = MqTransmitter(self.logger)
        self.timetable = timetable
        self.uow_dao = UnitOfWorkDao(self.logger)
        self.job_dao = JobDao(self.logger)

    @property
    def run_on_active_timeperiod(self):
        """
        :return: True if given State Machine allows execution on *live* timeperiod, as opposed to the finished one
        """
        raise NotImplementedError('property run_on_active_timeperiod must be implemented by {0}'
                                  .format(self.__class__.__name__))

    def _log_message(self, level, process_name, timeperiod, msg):
        """ method performs logging into log file and Timetable's tree node"""
        self.timetable.add_log_entry(process_name, timeperiod, msg)
        self.logger.log(level, msg)

    @with_reconnect
    def _insert_uow(self, process_name, timeperiod, start_timeperiod, end_timeperiod, start_id, end_id):
        """creates unit_of_work and inserts it into the DB
            :raise DuplicateKeyError: if unit_of_work with given parameters already exists """
        uow = UnitOfWork()
        uow.process_name = process_name
        uow.timeperiod = timeperiod
        uow.start_id = str(start_id)
        uow.end_id = str(end_id)
        uow.start_timeperiod = start_timeperiod
        uow.end_timeperiod = end_timeperiod
        uow.created_at = datetime.utcnow()
        uow.submitted_at = datetime.utcnow()
        uow.source = context.process_context[process_name].source
        uow.sink = context.process_context[process_name].sink
        uow.state = unit_of_work.STATE_REQUESTED
        uow.unit_of_work_type = unit_of_work.TYPE_MANAGED
        uow.number_of_retries = 0
        uow.arguments = context.process_context[process_name].arguments
        uow.db_id = self.uow_dao.insert(uow)

        msg = 'Created: UOW {0} for {1}@{2}.'.format(uow.db_id, process_name, start_timeperiod)
        self._log_message(INFO, process_name, start_timeperiod, msg)
        return uow

    def _publish_uow(self, uow):
        self.mq_transmitter.publish_managed_uow(uow)
        msg = 'Published: UOW {0} for {1}@{2}.'.format(uow.db_id, uow.process_name, uow.start_timeperiod)
        self._log_message(INFO, uow.process_name, uow.start_timeperiod, msg)

    def insert_and_publish_uow(self, job_record, start_id, end_id):
        """ method creates and publishes a unit_of_work. it also handles DuplicateKeyError and attempts recovery
        :return: tuple (uow, is_duplicate)
        :raise UserWarning: if the recovery from DuplicateKeyError was unsuccessful
        """
        process_name = job_record.process_name
        timeperiod = job_record.timeperiod
        start_timeperiod = self.compute_start_timeperiod(job_record.process_name, job_record.timeperiod)
        end_timeperiod = self.compute_end_timeperiod(job_record.process_name, job_record.timeperiod)

        try:
            is_duplicate = False
            uow = self._insert_uow(process_name, timeperiod, start_timeperiod, end_timeperiod, start_id, end_id)
        except DuplicateKeyError as e:
            is_duplicate = True
            msg = 'Catching up with latest UOW {0}@{1}, because of: {2}' \
                  .format(process_name, start_timeperiod, e)
            self._log_message(WARNING, process_name, start_timeperiod, msg)
            uow = self.uow_dao.recover_from_duplicatekeyerror(e)

        if not uow:
            msg = 'PERSISTENT TIER ERROR! Unable to locate UOW for {0}@{1}' \
                  .format(process_name, start_timeperiod)
            self._log_message(WARNING, process_name, start_timeperiod, msg)
            raise UserWarning(msg)

        if uow.is_canceled:
            # this UOW was marked for re-processing. recycle it
            uow.created_at = datetime.utcnow()      # reset created_at to bypass GC cancellation logic
            uow.submitted_at = datetime.utcnow()    # reset submitted_at to allow 1 hour free of GC resubmitting
            uow.state = unit_of_work.STATE_REQUESTED

        # publish the created/caught up unit_of_work
        self._publish_uow(uow)
        return uow, is_duplicate

    def notify(self, uow):
        """ method is used by StateMachine's users to provide notifications about change in UOW
        if applicable, method will update job_record state and Timetable tree node state
        :assumptions: uow is in [STATE_NOOP, STATE_CANCELED, STATE_PROCESSED] """
        pass

    def compute_start_timeperiod(self, process_name, timeperiod):
        """ computes lowest *inclusive* timeperiod boundary for job to process
            for process with time_grouping == 1, it returns given timeperiod with no change
            for process with time_grouping != 1, it computes first timeperiod, not processed by the previous job run
            For instance: with time_grouping = 3, QUALIFIER_HOURLY, and timeperiod = 2016042018,
            the start_timeperiod will be = 2016042016 (computed as 2016042018 - 3 + 1)
        """
        time_grouping = context.process_context[process_name].time_grouping
        if time_grouping == 1:
            return timeperiod

        # step1: translate given timeperiod to the time grouped one
        process_hierarchy = self.timetable.get_tree(process_name).process_hierarchy
        timeperiod_dict = process_hierarchy[process_name].timeperiod_dict
        translated_timeperiod = timeperiod_dict._translate_timeperiod(timeperiod)

        # step 2: compute previous grouped period
        # NOTICE: simple `time_helper.increment_timeperiod(time_qualifier, timeperiod)` is insufficient
        #         as it does not address edge cases, such as the last day of the month or the last hour of the day
        # For instance: with time_grouping=3, QUALIFIER_DAILY, and 2016123100
        # the `increment_timeperiod` will yield 2016122800 instead of 2016123100
        time_qualifier = context.process_context[process_name].time_qualifier
        for i in range(1, time_grouping + 1):
            prev_timeperiod = time_helper.increment_timeperiod(time_qualifier, translated_timeperiod, delta=-i)
            if prev_timeperiod == timeperiod_dict._translate_timeperiod(prev_timeperiod):
                # prev_timeperiod is currently at the last grouped timeperiod
                break

        # step 3: compute first exclusive timeperiod after the *prev_timeperiod*,
        # which becomes first inclusive timeperiod for this job run
        over_the_edge_timeperiod = time_helper.increment_timeperiod(time_qualifier, prev_timeperiod, delta=-1)
        if prev_timeperiod != timeperiod_dict._translate_timeperiod(over_the_edge_timeperiod):
            # over_the_edge_timeperiod fell into previous day or month or year
            # *prev_timeperiod* points to the first month, first day of the month or 00 hour
            start_timeperiod = prev_timeperiod
        else:
            start_timeperiod = self.compute_end_timeperiod(process_name, prev_timeperiod)

        return start_timeperiod

    def compute_end_timeperiod(self, process_name, timeperiod):
        """ computes first *exclusive* timeperiod for job to process """
        time_qualifier = context.process_context[process_name].time_qualifier
        end_timeperiod = time_helper.increment_timeperiod(time_qualifier, timeperiod)
        return end_timeperiod

    def _is_noop_timeperiod(self, process_name, timeperiod):
        """ method verifies if the given timeperiod for given process is valid or falls in-between grouping checkpoints
        :param process_name: name of the process
        :param timeperiod: timeperiod to verify
        :return: False, if given process has no time_grouping set or it is equal to 1.
                 False, if time_grouping is custom but the given timeperiod matches the grouped timeperiod.
                 True, if the timeperiod falls in-between grouping cracks
        """
        time_grouping = context.process_context[process_name].time_grouping
        if time_grouping == 1:
            return False

        process_hierarchy = self.timetable.get_tree(process_name).process_hierarchy
        timeperiod_dict = process_hierarchy[process_name].timeperiod_dict
        return timeperiod_dict._translate_timeperiod(timeperiod) != timeperiod

    def _process_noop_timeperiod(self, job_record):
        """ method is valid for processes having time_grouping != 1.
            should a job record fall in-between grouped time milestones,
            its state should be set to STATE_NOOP without any processing """
        job_record.state = job.STATE_NOOP
        self.job_dao.update(job_record)

        time_grouping = context.process_context[job_record.process_name].time_grouping
        msg = 'Job {0}@{1} with time_grouping {2} was transferred to STATE_NOOP' \
              .format(job_record.process_name, job_record.timeperiod, time_grouping)
        self._log_message(INFO, job_record.process_name, job_record.timeperiod, msg)

    def _process_state_embryo(self, job_record):
        """ method that takes care of processing job records in STATE_EMBRYO state"""
        pass

    def _process_state_in_progress(self, job_record):
        """ method that takes care of processing job records in STATE_IN_PROGRESS state"""
        pass

    def _process_state_final_run(self, job_record):
        """method takes care of processing job records in STATE_FINAL_RUN state"""
        uow = self.uow_dao.get_one(job_record.related_unit_of_work)
        if uow.is_processed:
            self.update_job(job_record, uow, job.STATE_PROCESSED)
        elif uow.is_noop:
            self.update_job(job_record, uow, job.STATE_NOOP)
        elif uow.is_canceled:
            self.update_job(job_record, uow, job.STATE_SKIPPED)
        elif uow.is_invalid:
            msg = 'Job {0}: UOW for {1}@{2} is in {3}; ' \
                  'relying on the Garbage Collector to either recycle or cancel the UOW.' \
                  .format(job_record.db_id, job_record.process_name, job_record.timeperiod, uow.state)
            self._log_message(INFO, job_record.process_name, job_record.timeperiod, msg)
        else:
            msg = 'Suppressed creating UOW for {0}@{1}; Job is in {2}; uow is in {3}' \
                  .format(job_record.process_name, job_record.timeperiod, job_record.state, uow.state)
            self._log_message(INFO, job_record.process_name, job_record.timeperiod, msg)

        timetable_tree = self.timetable.get_tree(job_record.process_name)
        timetable_tree.build_tree()

    def _process_terminal_state(self, job_record):
        """ method logs a warning message notifying that the job is no longer govern by this state machine """
        msg = 'Job {0} for {1}@{2} is in the terminal state {3}, ' \
              'and is no further govern by the State Machine {4}' \
              .format(job_record.db_id, job_record.process_name, job_record.timeperiod, job_record.state, self.name)
        self._log_message(WARNING, job_record.process_name, job_record.timeperiod, msg)

    def manage_job_with_blocking_children(self, job_record):
        """ method will trigger job processing only if:
            - all children are finished (STATE_PROCESSED, STATE_SKIPPED, STATE_NOOP)
            - all external dependencies are finalized (STATE_PROCESSED, STATE_SKIPPED, STATE_NOOP) """
        is_job_finalizable = self.timetable.is_job_record_finalizable(job_record)
        if is_job_finalizable:
            self.manage_job(job_record)
        else:
            msg = 'Job {0}@{1} is blocked by unprocessed children/dependencies. Waiting another tick' \
                  .format(job_record.process_name, job_record.timeperiod)
            self._log_message(INFO, job_record.process_name, job_record.timeperiod, msg)

    def manage_job_with_blocking_dependencies(self, job_record):
        """ method will trigger job processing only if:
            - all dependencies are in [STATE_PROCESSED, STATE_NOOP]
            NOTICE: method will transfer current job into STATE_SKIPPED if any dependency is in STATE_SKIPPED """
        composite_state = self.timetable.dependent_on_composite_state(job_record)
        assert isinstance(composite_state, NodesCompositeState)

        if composite_state.all_processed:
            self.manage_job(job_record)
        elif composite_state.skipped_present:
            # As soon as among <dependent on> periods are in STATE_SKIPPED
            # there is very little sense in waiting for them to become STATE_PROCESSED
            # Skip this timeperiod itself
            job_record.state = job.STATE_SKIPPED
            self.job_dao.update(job_record)
            self.mq_transmitter.publish_job_status(job_record)

            msg = 'Job {0}@{1} is blocked by STATE_SKIPPED dependencies. ' \
                  'Transferred the job to STATE_SKIPPED'.format(job_record.process_name, job_record.timeperiod)
            self._log_message(WARNING, job_record.process_name, job_record.timeperiod, msg)
        else:
            msg = 'Job {0}@{1} is blocked by unprocessed dependencies. Waiting another tick' \
                  .format(job_record.process_name, job_record.timeperiod)
            self._log_message(INFO, job_record.process_name, job_record.timeperiod, msg)

    def manage_job(self, job_record):
        """ method main duty - is to _avoid_ publishing another unit_of_work, if previous was not yet processed
            In case the State Machine sees that the unit_of_work already exist it could either:
            - update boundaries of the processing; republish
            - wait another tick; idle """
        assert isinstance(job_record, Job)
        try:
            if self._is_noop_timeperiod(job_record.process_name, job_record.timeperiod):
                self._process_noop_timeperiod(job_record)
                return

            if job_record.is_embryo:
                self._process_state_embryo(job_record)

            elif job_record.is_in_progress:
                self._process_state_in_progress(job_record)

            elif job_record.is_final_run:
                self._process_state_final_run(job_record)

            elif job_record.is_skipped:
                self._process_terminal_state(job_record)

            elif job_record.is_processed:
                self._process_terminal_state(job_record)

            elif job_record.is_noop:
                self._process_terminal_state(job_record)

            else:
                msg = 'Unknown state {0} of the job {1}'.format(job_record.state, job_record.db_id)
                self._log_message(ERROR, job_record.process_name, job_record.timeperiod, msg)

        except LookupError as e:
            job_record.number_of_failures += 1
            self.job_dao.update(job_record)
            self.timetable.skip_if_needed(job_record)
            msg = 'Increasing fail counter for Job {0}@{1}, because of: {2}' \
                  .format(job_record.process_name, job_record.timeperiod, e)
            self._log_message(WARNING, job_record.process_name, job_record.timeperiod, msg)

        finally:
            self.mq_transmitter.publish_job_status(job_record)

    def reprocess_job(self, job_record):
        """ method marks given job for reprocessing:
            - if no UOW was bind with the job record - it is transferred to STATE_EMBRYO only
            - if related UOW was not finished - we mark the UOW as STATE_CANCELED
                and pass the job record thru the _process_in_progress routine
            - otherwise job record is passed thru the _process_in_progress routine """
        original_job_state = job_record.state
        job_record.number_of_failures = 0

        if not job_record.related_unit_of_work:
            job_record.state = job.STATE_EMBRYO
            self.job_dao.update(job_record)
        else:
            uow = self.uow_dao.get_one(job_record.related_unit_of_work)
            if not uow.is_finished:
                uow.state = unit_of_work.STATE_CANCELED
                uow.submitted_at = datetime.utcnow()
                self.uow_dao.update(uow)
            self._process_state_in_progress(job_record)

        msg = 'Reprocessed Job {0} for {1}@{2}: state transfer {3} -> {4};' \
              .format(job_record.db_id, job_record.process_name, job_record.timeperiod,
                      original_job_state, job_record.state)
        self._log_message(WARNING, job_record.process_name, job_record.timeperiod, msg)

    def skip_job(self, job_record):
        """ method transfers:
            - given job into STATE_SKIPPED if it is not not in finished state
            - UOW into STATE_CANCELED if it is not in finished state """
        original_job_state = job_record.state

        if not job_record.is_finished:
            job_record.state = job.STATE_SKIPPED
            self.job_dao.update(job_record)

        if job_record.related_unit_of_work:
            uow = self.uow_dao.get_one(job_record.related_unit_of_work)
            if not uow.is_finished:
                uow.state = unit_of_work.STATE_CANCELED
                uow.submitted_at = datetime.utcnow()
                self.uow_dao.update(uow)

        msg = 'Skipped Job {0} for {1}@{2}: state transfer {3} -> {4};' \
              .format(job_record.db_id, job_record.process_name, job_record.timeperiod,
                      original_job_state, job_record.state)
        self._log_message(WARNING, job_record.process_name, job_record.timeperiod, msg)

    def create_job(self, process_name, timeperiod):
        """ method creates a job record in STATE_EMBRYO for given process_name and timeperiod
            :returns: created job record of type <Job>"""
        job_record = Job()
        job_record.state = job.STATE_EMBRYO
        job_record.timeperiod = timeperiod
        job_record.process_name = process_name
        self.job_dao.update(job_record)

        self.logger.info('Created Job {0} for {1}@{2}'
                         .format(job_record.db_id, job_record.process_name, job_record.timeperiod))
        return job_record

    def update_job(self, job_record, uow, new_state):
        """ method updates job record with a new unit_of_work and new state"""
        original_job_state = job_record.state
        job_record.state = new_state
        job_record.related_unit_of_work = uow.db_id
        self.job_dao.update(job_record)

        msg = 'Updated Job {0} for {1}@{2}: state transfer {3} -> {4};' \
              .format(job_record.db_id, job_record.process_name, job_record.timeperiod, original_job_state, new_state)
        self._log_message(INFO, job_record.process_name, job_record.timeperiod, msg)



__author__ = 'Bohdan Mushkevych'

from threading import Thread
import socket

from amqp import AMQPError

from synergy.conf import settings
from synergy.mq.flopsy import Consumer
from synergy.system.performance_tracker import SimpleTracker
from synergy.system.synergy_process import SynergyProcess


class AbstractMqWorker(SynergyProcess):
    """
    class works as an abstract basement for all workers and aggregators
    it registers in the mq and awaits for the messages
    """

    def __init__(self, process_name):
        """:param process_name: id of the process, the worker will be performing """
        super(AbstractMqWorker, self).__init__(process_name)
        self.mq_timeout_seconds = 0
        self._init_mq_timeout_seconds()

        self.consumer = None
        self._init_mq_consumer()

        self.main_thread = None
        self.performance_tracker = None
        self._init_performance_tracker(self.logger)

        msg_suffix = 'in Production Mode'
        if settings.settings['under_test']:
            msg_suffix = 'in Testing Mode'
        self.logger.info('Started {0} {1}'.format(self.process_name, msg_suffix))

    def __del__(self):
        try:
            self.logger.info('Closing Flopsy Consumer...')
            self.consumer.close()
        except Exception as e:
            self.logger.error('Exception caught while closing Flopsy Consumer: {0}'.format(e))

        try:
            self.logger.info('Canceling Performance Tracker...')
            self.performance_tracker.cancel()
        except Exception as e:
            self.logger.error('Exception caught while cancelling the Performance Tracker: {0}'.format(e))
        super(AbstractMqWorker, self).__del__()

    # ********************** abstract methods ****************************
    def _init_performance_tracker(self, logger):
        self.performance_tracker = SimpleTracker(logger)
        self.performance_tracker.start()

    def _init_mq_consumer(self):
        self.consumer = Consumer(self.process_name)

    def _init_mq_timeout_seconds(self):
        if 'mq_timeout_sec' in settings.settings:
            self.mq_timeout_seconds = settings.settings['mq_timeout_sec']

    # ********************** thread-related methods ****************************
    def _mq_callback(self, message):
        """ abstract method to process messages from MQ
        :param message: mq message"""
        pass

    def _run_mq_listener(self):
        try:
            self.consumer.register(self._mq_callback)
            self.consumer.wait(self.mq_timeout_seconds)
        except socket.timeout as e:
            self.logger.warning('Queue {0} is likely empty. Worker exits due to: {1}'.format(self.consumer.queue, e))
        except (AMQPError, IOError) as e:
            self.logger.error('AMQPError: {0}'.format(e))
        finally:
            self.__del__()
            self.logger.info('Exiting main thread. All auxiliary threads stopped.')

    def start(self, *_):
        self.main_thread = Thread(target=self._run_mq_listener)
        self.main_thread.start()



__author__ = 'Bohdan Mushkevych'

from synergy.system.time_trigger_factory import format_time_trigger_string
from synergy.mx.rest_model import *


def get_next_run_in(thread_handler):
    if not thread_handler.is_alive:
        return 'NA'

    next_run = thread_handler.next_run_in()
    return str(next_run).split('.')[0]


def get_next_timeperiod(timetable, process_name):
    if timetable.get_tree(process_name) is None:
        return 'NA'
    else:
        job_record = timetable.get_next_job_record(process_name)
        return job_record.timeperiod


def get_dependant_trees(timetable, tree_obj):
    trees = timetable._find_dependant_trees(tree_obj)
    return [x.tree_name for x in trees]


def get_reprocessing_queue(gc, process_name):
    per_process = gc.reprocess_uows[process_name]
    q = []
    for priority_entry in sorted(per_process.queue):
        q.append(priority_entry.entry.timeperiod)
    return q


def create_rest_managed_scheduler_entry(thread_handler, timetable, gc):
    process_entry = thread_handler.process_entry
    process_name = process_entry.process_name

    rest_model = RestManagedSchedulerEntry(
        is_on=process_entry.is_on,
        is_alive=thread_handler.is_alive,
        process_name=process_name,
        trigger_frequency=format_time_trigger_string(thread_handler.timer_instance),
        next_run_in=get_next_run_in(thread_handler),
        next_timeperiod=get_next_timeperiod(timetable, process_name),
        time_qualifier=process_entry.time_qualifier,
        time_grouping=process_entry.time_grouping,
        state_machine_name=process_entry.state_machine_name,
        blocking_type=process_entry.blocking_type,
        reprocessing_queue=get_reprocessing_queue(gc, process_name),
    )
    return rest_model


def create_rest_freerun_scheduler_entry(thread_handler):
    process_name, entry_name = thread_handler.key
    rest_model = RestFreerunSchedulerEntry(
        is_on=thread_handler.process_entry.is_on,
        is_alive=thread_handler.is_alive,
        process_name=process_name,
        entry_name=entry_name,
        trigger_frequency=format_time_trigger_string(thread_handler.timer_instance),
        description=thread_handler.process_entry.description,
        next_run_in=get_next_run_in(thread_handler),
        event_log=thread_handler.process_entry.event_log,
        arguments=thread_handler.process_entry.arguments
    )
    return rest_model


def create_rest_timetable_tree(timetable, tree_obj):
    rest_tree = RestTimetableTree(tree_name=tree_obj.tree_name,
                                  mx_page=tree_obj.mx_page,
                                  mx_name=tree_obj.mx_name,
                                  dependent_on=[tree.tree_name for tree in tree_obj.dependent_on],
                                  dependant_trees=get_dependant_trees(timetable, tree_obj),
                                  sorted_process_names=[x for x in tree_obj.process_hierarchy])
    return rest_tree



__author__ = 'Bohdan Mushkevych'

import collections
from threading import Lock
from datetime import datetime, timedelta

from synergy.conf import settings
from synergy.system.system_logger import get_logger
from synergy.system.decorator import thread_safe
from synergy.system.priority_queue import PriorityEntry, PriorityQueue, compute_release_time
from synergy.system.repeat_timer import RepeatTimer
from synergy.system.mq_transmitter import MqTransmitter
from synergy.scheduler.scheduler_constants import PROCESS_GC
from synergy.scheduler.thread_handler import ManagedThreadHandler
from synergy.db.model import unit_of_work
from synergy.db.dao.unit_of_work_dao import UnitOfWorkDao


class GarbageCollector(object):
    """ GC is triggered directly by Synergy Scheduler.
        It scans for invalid or stalled unit_of_work and re-triggers them.
        GC is vital for the health of the system.
        Deployment with no GC is considered invalid """

    def __init__(self, scheduler):
        self.logger = get_logger(PROCESS_GC, append_to_console=False, redirect_stdstream=False)
        self.managed_handlers = scheduler.managed_handlers
        self.mq_transmitter = MqTransmitter(self.logger)
        self.timetable = scheduler.timetable

        self.lock = Lock()
        self.uow_dao = UnitOfWorkDao(self.logger)
        self.reprocess_uows = collections.defaultdict(PriorityQueue)
        self.timer = RepeatTimer(settings.settings['gc_run_interval'], self._run)

    @thread_safe
    def scan_uow_candidates(self):
        """ method performs two actions:
            - enlist stale or invalid units of work into reprocessing queue
            - cancel UOWs that are older than 2 days and have been submitted more than 1 hour ago """
        try:
            since = settings.settings['synergy_start_timeperiod']
            uow_list = self.uow_dao.get_reprocessing_candidates(since)
        except LookupError as e:
            self.logger.info('flow: no UOW candidates found for reprocessing: {0}'.format(e))
            return

        for uow in uow_list:
            try:
                if uow.process_name not in self.managed_handlers:
                    self.logger.debug('process {0} is not known to the Synergy Scheduler. Skipping its UOW.'
                                      .format(uow.process_name))
                    continue

                thread_handler = self.managed_handlers[uow.process_name]
                assert isinstance(thread_handler, ManagedThreadHandler)
                if not thread_handler.process_entry.is_on:
                    self.logger.debug('process {0} is inactive. Skipping its UOW.'.format(uow.process_name))
                    continue

                entry = PriorityEntry(uow)
                if entry in self.reprocess_uows[uow.process_name]:
                    # given UOW is already registered in the reprocessing queue
                    continue

                # ASSUMPTION: UOW is re-created by a state machine during reprocessing
                # thus - any UOW older 2 days could be marked as STATE_CANCELED
                if datetime.utcnow() - uow.created_at > timedelta(hours=settings.settings['gc_life_support_hours']):
                    self._cancel_uow(uow)
                    continue

                # if the UOW has been idle for more than 1 hour - resubmit it
                if datetime.utcnow() - uow.submitted_at > timedelta(hours=settings.settings['gc_resubmit_after_hours'])\
                        or uow.is_invalid:
                    # enlist the UOW into the reprocessing queue
                    self.reprocess_uows[uow.process_name].put(entry)

            except Exception as e:
                self.logger.error('flow exception: {0}'.format(e), exc_info=True)

    def _flush_queue(self, q, ignore_priority=False):
        """
        :param q: PriorityQueue instance holding GarbageCollector entries
        :param ignore_priority: If True - all GarbageCollector entries should be resubmitted
                If False - only those entries whose waiting time has expired will be resubmitted
        """
        assert isinstance(q, PriorityQueue)

        current_timestamp = compute_release_time(lag_in_minutes=0)
        for _ in range(len(q)):
            entry = q.pop()
            assert isinstance(entry, PriorityEntry)

            if ignore_priority or entry.release_time < current_timestamp:
                self._resubmit_uow(entry.entry)
            else:
                q.put(entry)
                break

    @thread_safe
    def flush(self, ignore_priority=False):
        """ method iterates over each reprocessing queues and re-submits UOW whose waiting time has expired """
        for process_name, q in self.reprocess_uows.items():
            self._flush_queue(q, ignore_priority)

    @thread_safe
    def validate(self):
        """ method iterates over the reprocessing queue and synchronizes state of every UOW with the DB
            should it change via the MX to STATE_CANCELED - remove the UOW from the queue """
        for process_name, q in self.reprocess_uows.items():
            if not q:
                continue

            invalid_entries = list()
            for entry in q.queue:
                assert isinstance(entry, PriorityEntry)
                uow = self.uow_dao.get_one(entry.entry.db_id)
                if uow.is_canceled:
                    invalid_entries.append(entry)

                thread_handler = self.managed_handlers[uow.process_name]
                assert isinstance(thread_handler, ManagedThreadHandler)
                if not thread_handler.process_entry.is_on:
                    invalid_entries.append(entry)

            for entry in invalid_entries:
                q.queue.remove(entry)

        self.logger.info('reprocessing queue validated')

    @thread_safe
    def flush_one(self, process_name, ignore_priority=False):
        """ method iterates over the reprocessing queue for the given process
            and re-submits UOW whose waiting time has expired """
        q = self.reprocess_uows[process_name]
        self._flush_queue(q, ignore_priority)

    def _resubmit_uow(self, uow):
        # re-read UOW from the DB, in case it was STATE_CANCELLED by MX
        uow = self.uow_dao.get_one(uow.db_id)
        if uow.is_canceled:
            self.logger.info('suppressed re-submission of UOW {0} for {1}@{2} in {3};'
                             .format(uow.db_id, uow.process_name, uow.timeperiod, uow.state))
            return

        thread_handler = self.managed_handlers[uow.process_name]
        assert isinstance(thread_handler, ManagedThreadHandler)
        if not thread_handler.process_entry.is_on:
            self.logger.debug('suppressed re-submission of UOW {0} for {1}@{2} in {3}, since the process is inactive.'
                              .format(uow.db_id, uow.process_name, uow.timeperiod, uow.state))
            return

        if uow.is_invalid:
            uow.number_of_retries += 1

        uow.state = unit_of_work.STATE_REQUESTED
        uow.submitted_at = datetime.utcnow()
        self.uow_dao.update(uow)

        self.mq_transmitter.publish_managed_uow(uow)
        self.logger.info('re-submitted UOW {0} for {1}@{2}; attempt {3}'
                         .format(uow.db_id, uow.process_name, uow.timeperiod, uow.number_of_retries))

    def _cancel_uow(self, uow):
        uow.state = unit_of_work.STATE_CANCELED
        self.uow_dao.update(uow)

        self.mq_transmitter.publish_uow_status(uow)
        self.logger.info('canceled UOW {0} for {1}@{2}; attempt {3}; created at {4}'
                         .format(uow.db_id, uow.process_name, uow.timeperiod, uow.number_of_retries, uow.created_at))

    def _run(self):
        try:
            self.logger.info('run {')

            self.logger.debug('step 1: validate existing queue entries')
            self.validate()

            self.logger.debug('step 2: scan reprocessing candidates')
            self.scan_uow_candidates()

            self.logger.debug('step 3: repost after timeout')
            self.flush()

            self.logger.debug('step 4: timetable housekeeping')
            self.timetable.build_trees()

            self.logger.debug('step 5: timetable validation')
            self.timetable.validate()
        except Exception as e:
            self.logger.error('GC run exception: {0}'.format(e))
        finally:
            self.logger.info('}')

    def start(self):
        self.timer.start()

    def stop(self):
        self.timer.cancel()



__author__ = 'Bohdan Mushkevych'

import json
from os import path
from datetime import datetime

from jinja2 import Environment, FileSystemLoader
from werkzeug.local import Local, LocalManager
from werkzeug.wrappers import Response
from werkzeug.routing import Map, Rule

from synergy.conf import context
from synergy.conf import settings

from flow.mx import STATIC_FLOW_ENDPOINT, STATIC_FLOW_PATH

TEMPLATE_PATH = path.join(path.dirname(__file__), 'templates')
STATIC_PATH = path.join(path.dirname(__file__), 'static')

local = Local()
local_manager = LocalManager([local])

# Synergy MX map of URL routing
url_map = Map()
url_map.add(Rule('/static/<file>', endpoint='static', build_only=True))
url_map.add(Rule('/{0}/<file>'.format(STATIC_FLOW_ENDPOINT), endpoint=STATIC_FLOW_ENDPOINT, build_only=True))

# tree/group of trees will be shown on a separate page defined by tree property MX_PAGE
# mx_page_context is a dictionary in format: {MX_PAGE: MX PAGE}
mx_page_context = {tree_entry.mx_page: tree_entry.mx_page.replace('_', ' ')
                   for tree_entry in context.timetable_context.values()}

# loop sets a Rule per every mx_page from mx_page_context to be processed by
# 'mx_page_tiles' method from mx.views.py
# NOTICE: given renders template snippet {{ url_for ('function_name') }} invalid,
# since all mx_page are processed by the single function 'mx_page_tiles'
for rule in mx_page_context:
    url_map.add(Rule('/{0}/'.format(rule), endpoint='mx_page_tiles'))


def expose(rule, **kw):
    def decorate(f):
        url_map.add(Rule(rule, endpoint=f.__name__))
        return f
    return decorate


def url_for(endpoint, _external=False, **values):
    return local.url_adapter.build(endpoint, values, force_external=_external)


def render_template(template, **context):
    return Response(jinja_env.get_template(template).render(**context), mimetype='text/html')


def scheduler_uptime():
    time_diff = datetime.utcnow() - settings.settings['process_start_time']
    d = {'days': time_diff.days}
    d['hours'], rem = divmod(time_diff.seconds, 3600)
    d['minutes'], d['seconds'] = divmod(rem, 60)
    return '{days:02d}:{hours:02d}:{minutes:02d}:{seconds:02d}'.format(**d)


jinja_env = Environment(loader=FileSystemLoader([TEMPLATE_PATH, STATIC_FLOW_PATH]), autoescape=True)
jinja_env.add_extension('jinja2.ext.do')
jinja_env.globals['url_for'] = url_for
jinja_env.globals['local'] = local
jinja_env.globals['get_current_time'] = lambda: datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S %Z')
jinja_env.globals['scheduler_version'] = lambda: settings.settings['version']
jinja_env.globals['scheduler_uptime'] = scheduler_uptime
jinja_env.globals['mx_processing_context'] = mx_page_context
jinja_env.globals['synergy_process_context'] = context.process_context
jinja_env.filters['jsonify'] = json.dumps



__author__ = 'Bohdan Mushkevych'

from threading import Thread

from amqp import AMQPError

from synergy.db.model.mq_transmission import MqTransmission
from synergy.db.model import unit_of_work
from synergy.db.dao.unit_of_work_dao import UnitOfWorkDao
from synergy.scheduler.scheduler_constants import QUEUE_UOW_STATUS
from synergy.mq.flopsy import Consumer


class UowStatusListener(object):
    """ class instance listens to the QUEUE_UOW_STATUS queue and updates Timetable records correspondingly """

    def __init__(self, scheduler):
        self.scheduler = scheduler
        self.timetable = scheduler.timetable
        self.logger = scheduler.logger
        self.uow_dao = UnitOfWorkDao(self.logger)
        self.consumer = Consumer(QUEUE_UOW_STATUS)
        self.main_thread = None

    def __del__(self):
        try:
            self.logger.info('UowStatusListener: Closing Flopsy Consumer...')
            self.consumer.close()
        except Exception as e:
            self.logger.error('UowStatusListener: Exception caught while closing Flopsy Consumer: {0}'.format(e))

    # ********************** thread-related methods ****************************
    def _mq_callback(self, message):
        """ method processes messages from Synergy Worker and updates corresponding Timetable record,
            as well as the job itself
            :param message: <MqTransmission> mq message """
        try:
            self.logger.info('UowStatusListener {')

            mq_request = MqTransmission.from_json(message.body)
            uow = self.uow_dao.get_one(mq_request.record_db_id)
            if uow.unit_of_work_type != unit_of_work.TYPE_MANAGED:
                self.logger.info('Received transmission from non-managed UOW execution: {0}. Ignoring it.'
                                 .format(uow.unit_of_work_type))
                return

            tree = self.timetable.get_tree(uow.process_name)
            node = tree.get_node(uow.process_name, uow.timeperiod)

            if uow.db_id != node.job_record.related_unit_of_work:
                self.logger.info('Received transmission is likely outdated. Ignoring it.')
                return

            if not uow.is_finished:
                # rely on Garbage Collector to re-trigger the failing unit_of_work
                self.logger.info('Received transmission from {0}@{1} in non-final state {2}. Ignoring it.'
                                 .format(uow.process_name, uow.timeperiod, uow.state))
                return

            state_machine = self.scheduler.state_machine_for(node.process_name)
            self.logger.info('Commencing StateMachine.notify with UOW from {0}@{1} in {2}.'
                             .format(uow.process_name, uow.timeperiod, uow.state))
            state_machine.notify(uow)

        except KeyError:
            self.logger.error('Access error for {0}'.format(message.body), exc_info=True)
        except Exception:
            self.logger.error('Error during StateMachine.notify call {0}'.format(message.body), exc_info=True)
        finally:
            self.consumer.acknowledge(message.delivery_tag)
            self.logger.info('UowStatusListener }')

    def _run_mq_listener(self):
        try:
            self.consumer.register(self._mq_callback)
            self.logger.info('UowStatusListener: instantiated and activated.')
            self.consumer.wait()
        except (AMQPError, IOError) as e:
            self.logger.error('UowStatusListener: AMQPError {0}'.format(e))
        finally:
            self.__del__()
            self.logger.info('UowStatusListener: Shut down.')

    def start(self, *_):
        self.main_thread = Thread(target=self._run_mq_listener)
        self.main_thread.daemon = True
        self.main_thread.start()

    def stop(self):
        """ method stops currently MQ Consumer listener, if any """
        self.__del__()



__author__ = 'Bohdan Mushkevych'

import logging
from datetime import datetime

from synergy.db.model.log_recording import LogRecording
from synergy.db.dao.log_recording_dao import LogRecordingDao


class LogRecordingHandler(logging.Handler):
    def __init__(self, logger, parent_object_id):
        super(LogRecordingHandler, self).__init__()
        self.logger = logger
        self.parent_object_id = parent_object_id
        self.log_recording_dao = LogRecordingDao(logger)

    def attach(self):
        """ method clears existing log_recorder entries for given parent_object_id,
            creates a new one and attaches this handler to the logger
            from this moment every log record will be recorded in the DB """
        log_recording = LogRecording(parent_object_id=self.parent_object_id, created_at=datetime.utcnow())
        self.log_recording_dao.remove(self.parent_object_id)
        self.log_recording_dao.update(log_recording)

        formatter = logging.Formatter(fmt='%(asctime)s %(levelname)-8s %(message)s',
                                      datefmt='%Y-%m-%d %H:%M:%S')
        self.setFormatter(formatter)
        self.logger.addHandler(self)

    def detach(self):
        """ method detaches this handler from the logger """
        self.logger.removeHandler(self)

    def emit(self, record):
        msg = self.format(record)
        try:
            self.log_recording_dao.append_log(self.parent_object_id, msg.rstrip())
        except Exception as e:
            self.detach()
            self.logger.error('Detached LogRecordingHandler. Exception on LogRecordingDao.append_log: {0}'.format(e),
                              exc_info=True)



__author__ = 'Bohdan Mushkevych'

from synergy.db.dao.base_dao import BaseDao
from synergy.db.model.log_recording import LogRecording, PARENT_OBJECT_ID, LOG
from synergy.system.decorator import thread_safe
from synergy.scheduler.scheduler_constants import COLLECTION_LOG_RECORDING


class LogRecordingDao(BaseDao):
    """ Thread-safe Data Access Object for logs stored in log_recording table/collection """

    def __init__(self, logger):
        super(LogRecordingDao, self).__init__(logger=logger,
                                              model_class=LogRecording,
                                              primary_key=[PARENT_OBJECT_ID],
                                              collection_name=COLLECTION_LOG_RECORDING)

    @thread_safe
    def append_log(self, uow_id, msg):
        collection = self.ds.connection(self.collection_name)

        result = collection.update_one({PARENT_OBJECT_ID: uow_id},
                                       {'$push': {LOG: msg}},
                                       upsert=True)
        if result.modified_count == 0:
            raise LookupError('Log append failed for {0} in collection {1}'.format(uow_id, self.collection_name))



__author__ = 'Bohdan Mushkevych'

import sys
import time
import functools
import traceback


def thread_safe(method):
    """ wraps method with lock acquire/release cycle
     decorator requires class instance to have field self.lock of type threading.Lock or threading.RLock """

    @functools.wraps(method)
    def _locker(self, *args, **kwargs):
        assert hasattr(self, 'lock'), \
            'thread_safe decorator applied to method {0}.{1}: missing required field {0}.lock'.\
            format(self.__class__.__name__, method.__name__)

        try:
            self.lock.acquire()
            return method(self, *args, **kwargs)
        finally:
            try:
                self.lock.release()
            except:
                sys.stderr.write('Exception on releasing lock at method {0}'.format(method.__name__))
                traceback.print_exc(file=sys.stderr)

    return _locker


def with_reconnect(func):
    """
    Handle when AutoReconnect is raised from pymongo. This is the standard error
    raised for everything from "host disconnected" to "couldn't connect to host"
    and more.

    The sleep handles the edge case when the state of a replica set changes, and
    the cursor raises AutoReconnect because the master may have changed. It can
    take some time for the replica set to stop raising this exception, and the
    small sleep and iteration count gives us a couple of seconds before we fail
    completely.
    """
    from pymongo.errors import AutoReconnect

    @functools.wraps(func)
    def _reconnector(*args, **kwargs):
        for _ in range(20):
            try:
                return func(*args, **kwargs)
            except AutoReconnect:
                time.sleep(0.250)
        raise

    return _reconnector


def singleton(cls):
    """
    turns class to singleton
    :param cls: class itself
    :return: function that either creates new instance of the class or returns existing one
    """

    # the only way to implement nonlocal closure variables in Python 2.X
    instances = {}

    def get_instance():
        if cls not in instances:
            instances[cls] = cls()
        return instances[cls]

    return get_instance



__author__ = 'Bohdan Mushkevych'

from synergy.scheduler.scheduler_constants import STATE_MACHINE_FREERUN, EXCHANGE_FREERUN_WORKER
from synergy.db.model.daemon_process_entry import DaemonProcessEntry
from odm.fields import StringField, ListField, ObjectIdField, BooleanField

PROCESS_NAME = 'process_name'           # name of the process to handle the schedulables
ENTRY_NAME = 'entry_name'               # name of the schedulable
DESCRIPTION = 'description'             # description of the schedulable
IS_ON = 'is_on'                         # defines if the schedulable is active or off
TRIGGER_FREQUENCY = 'trigger_frequency'  # either 'at DoW-HH:MM' or 'every XXX'
STATE_MACHINE_NAME = 'state_machine_name'
SOURCE = 'source'
SINK = 'sink'

# contains list of last MAX_NUMBER_OF_EVENTS job events, such as emission of the UOW
EVENT_LOG = 'event_log'
MAX_NUMBER_OF_EVENTS = 128
RELATED_UNIT_OF_WORK = 'related_unit_of_work'


class FreerunProcessEntry(DaemonProcessEntry):
    """ Class presents single configuration entry for the freerun process/bash_driver """

    db_id = ObjectIdField('_id', null=True)
    source = StringField(SOURCE, null=True)
    sink = StringField(SINK, null=True)
    trigger_frequency = StringField(TRIGGER_FREQUENCY)
    is_on = BooleanField(IS_ON, default=False)
    state_machine_name = StringField(STATE_MACHINE_NAME)

    entry_name = StringField(ENTRY_NAME)
    description = StringField(DESCRIPTION)
    event_log = ListField(EVENT_LOG)
    related_unit_of_work = ObjectIdField(RELATED_UNIT_OF_WORK)

    @DaemonProcessEntry.key.getter
    def key(self):
        return self.process_name, self.entry_name

    @DaemonProcessEntry.key.setter
    def key(self, value):
        self.process_name = value[0]
        self.entry_name = value[1]

    @property
    def schedulable_name(self):
        return '{0}::{1}'.format(self.process_name, self.entry_name)


def freerun_context_entry(process_name,
                          entry_name,
                          classname,
                          token,
                          trigger_frequency,
                          is_on=True,
                          present_on_boxes=None,
                          description=None,
                          arguments=None,
                          exchange=EXCHANGE_FREERUN_WORKER,
                          queue=None,
                          routing=None,
                          pid_file=None,
                          log_file=None):
    """ forms process context entry """
    _ROUTING_PREFIX = 'routing_'
    _QUEUE_PREFIX = 'queue_'
    _SUFFIX = '_freerun'

    if arguments is not None:
        assert isinstance(arguments, dict)

    process_entry = FreerunProcessEntry(
        process_name=process_name,
        entry_name=entry_name,
        trigger_frequency=trigger_frequency,
        state_machine_name=STATE_MACHINE_FREERUN,
        is_on=is_on,
        classname=classname,
        token=token,
        present_on_boxes=present_on_boxes,
        description=description,
        mq_queue=queue if queue is not None else _QUEUE_PREFIX + token + _SUFFIX,
        mq_routing_key=routing if routing is not None else _ROUTING_PREFIX + token + _SUFFIX,
        mq_exchange=exchange,
        arguments=arguments if arguments is not None else dict(),
        log_filename=log_file if log_file is not None else token + _SUFFIX + '.log',
        pid_filename=pid_file if pid_file is not None else token + _SUFFIX + '.pid')
    return process_entry



__author__ = 'Bohdan Mushkevych'

from odm.document import BaseDocument
from odm.fields import StringField, ObjectIdField, IntegerField, DictField, DateTimeField

TIMEPERIOD = 'timeperiod'
START_TIMEPERIOD = 'start_timeperiod'  # lower boundary (as Synergy date) of the period that needs to be processed
END_TIMEPERIOD = 'end_timeperiod'      # upper boundary (as Synergy date) of the period that needs to be processed
START_ID = 'start_obj_id'              # lower boundary (as DB _id) of the period that needs to be processed
END_ID = 'end_obj_id'                  # upper boundary (as DB _id) of the period that needs to be processed
STATE = 'state'
CREATED_AT = 'created_at'
SUBMITTED_AT = 'submitted_at'
STARTED_AT = 'started_at'
FINISHED_AT = 'finished_at'
NUMBER_OF_AGGREGATED_DOCUMENTS = 'number_of_aggregated_documents'
NUMBER_OF_PROCESSED_DOCUMENTS = 'number_of_processed_documents'
NUMBER_OF_RETRIES = 'number_of_retries'

PROCESS_NAME = 'process_name'          # process name of the aggregator/alarm/etc that processed the range
SOURCE = 'source'                      # defines source of data for the computation
SINK = 'sink'                          # defines sink where the aggregated data will be inserted
ARGUMENTS = 'arguments'                # task-level arguments that could supplement or override process-level ones

UNIT_OF_WORK_TYPE = 'unit_of_work_type'     # whether the unit_of_work is TYPE_MANAGED or TYPE_FREERUN
TYPE_MANAGED = 'type_managed'               # identifies UOW created by Abstract State Machine child for Managed Process
TYPE_FREERUN = 'type_freerun'               # identifies UOW created by FreerunStateMachine for ad-hock processing

# UOW was successfully processed by the worker
STATE_PROCESSED = 'state_processed'

# UOW was received by the worker and it started the processing
STATE_IN_PROGRESS = 'state_in_progress'

# UOW was instantiated and send to the worker
STATE_REQUESTED = 'state_requested'

# Job has been manually marked as SKIPPED via MX
# and so the associated UOW got cancelled
# or the life-support threshold has been crossed for failing UOW
STATE_CANCELED = 'state_canceled'

# UOW can get into STATE_INVALID if:
# a. related Job was marked for reprocessing via MX
# b. have failed with an exception at the worker level
# NOTICE: GarbageCollector changes STATE_INVALID -> STATE_REQUESTED during re-posting
STATE_INVALID = 'state_invalid'

# UOW was received by a worker,
# but no data was found to process
STATE_NOOP = 'state_noop'


class UnitOfWork(BaseDocument):
    """ Module represents persistent Model for atomic unit of work performed by the system.
    UnitOfWork Instances are stored in the <unit_of_work> collection """

    db_id = ObjectIdField('_id', null=True)
    process_name = StringField(PROCESS_NAME)
    timeperiod = StringField(TIMEPERIOD, null=True)
    start_timeperiod = StringField(START_TIMEPERIOD, null=True)
    end_timeperiod = StringField(END_TIMEPERIOD, null=True)
    start_id = ObjectIdField(START_ID)
    end_id = ObjectIdField(END_ID)
    source = StringField(SOURCE, null=True)
    sink = StringField(SINK, null=True)
    arguments = DictField(ARGUMENTS)
    state = StringField(STATE, choices=[STATE_INVALID, STATE_REQUESTED, STATE_IN_PROGRESS,
                                        STATE_PROCESSED, STATE_CANCELED, STATE_NOOP])
    created_at = DateTimeField(CREATED_AT)
    submitted_at = DateTimeField(SUBMITTED_AT)
    started_at = DateTimeField(STARTED_AT)
    finished_at = DateTimeField(FINISHED_AT)

    number_of_aggregated_documents = IntegerField(NUMBER_OF_AGGREGATED_DOCUMENTS)
    number_of_processed_documents = IntegerField(NUMBER_OF_PROCESSED_DOCUMENTS)
    number_of_retries = IntegerField(NUMBER_OF_RETRIES, default=0)
    unit_of_work_type = StringField(UNIT_OF_WORK_TYPE, choices=[TYPE_MANAGED, TYPE_FREERUN])

    @property
    def key(self):
        return self.process_name, self.timeperiod, self.start_id, self.end_id

    @property
    def is_active(self):
        return self.state in [STATE_REQUESTED, STATE_IN_PROGRESS, STATE_INVALID]

    @property
    def is_finished(self):
        return self.state in [STATE_PROCESSED, STATE_CANCELED, STATE_NOOP]

    @property
    def is_processed(self):
        return self.state == STATE_PROCESSED

    @property
    def is_noop(self):
        return self.state == STATE_NOOP

    @property
    def is_canceled(self):
        return self.state == STATE_CANCELED

    @property
    def is_invalid(self):
        return self.state == STATE_INVALID

    @property
    def is_requested(self):
        return self.state == STATE_REQUESTED

    @property
    def is_in_progress(self):
        return self.state == STATE_IN_PROGRESS



__author__ = 'Bohdan Mushkevych'

from datetime import datetime
from logging import ERROR, WARNING, INFO

from synergy.conf import context
from synergy.db.dao.freerun_process_dao import FreerunProcessDao
from synergy.db.dao.unit_of_work_dao import UnitOfWorkDao
from synergy.db.error import DuplicateKeyError
from synergy.db.model import unit_of_work
from synergy.db.model.freerun_process_entry import FreerunProcessEntry, MAX_NUMBER_OF_EVENTS
from synergy.db.model.unit_of_work import UnitOfWork
from synergy.scheduler.scheduler_constants import STATE_MACHINE_FREERUN
from synergy.system import time_helper
from synergy.system.decorator import with_reconnect
from synergy.system.mq_transmitter import MqTransmitter
from synergy.system.time_qualifier import QUALIFIER_REAL_TIME


class StateMachineFreerun(object):
    """ State Machine to handle freerun jobs/triggers """

    def __init__(self, logger, name=STATE_MACHINE_FREERUN):
        self.name = name
        self.logger = logger
        self.mq_transmitter = MqTransmitter(self.logger)
        self.uow_dao = UnitOfWorkDao(self.logger)
        self.sfe_dao = FreerunProcessDao(self.logger)

    @with_reconnect
    def _log_message(self, level, freerun_entry, msg):
        """ method performs logging into log file and the freerun_entry """
        self.logger.log(level, msg)

        assert isinstance(freerun_entry, FreerunProcessEntry)
        event_log = freerun_entry.event_log
        if len(event_log) > MAX_NUMBER_OF_EVENTS:
            del event_log[-1]
        event_log.insert(0, msg)
        self.sfe_dao.update(freerun_entry)

    @with_reconnect
    def _insert_uow(self, freerun_entry, flow_request=None):
        """ creates unit_of_work and inserts it into the DB
            :raise DuplicateKeyError: if unit_of_work with given parameters already exists """
        process_entry = context.process_context[freerun_entry.process_name]
        arguments = process_entry.arguments
        arguments.update(freerun_entry.arguments)

        if flow_request:
            schedulable_name = flow_request.schedulable_name
            timeperiod = flow_request.timeperiod
            start_timeperiod = flow_request.start_timeperiod
            end_timeperiod = flow_request.end_timeperiod
            arguments.update(flow_request.arguments)
        else:
            schedulable_name = freerun_entry.schedulable_name
            timeperiod = time_helper.actual_timeperiod(QUALIFIER_REAL_TIME)
            start_timeperiod = timeperiod
            end_timeperiod = timeperiod

        uow = UnitOfWork()
        uow.process_name = schedulable_name
        uow.timeperiod = timeperiod
        uow.start_id = 0
        uow.end_id = 0
        uow.start_timeperiod = start_timeperiod
        uow.end_timeperiod = end_timeperiod
        uow.created_at = datetime.utcnow()
        uow.submitted_at = datetime.utcnow()
        uow.source = process_entry.source if hasattr(process_entry, 'source') else None
        uow.sink = process_entry.sink if hasattr(process_entry, 'sink') else None
        uow.state = unit_of_work.STATE_REQUESTED
        uow.unit_of_work_type = unit_of_work.TYPE_FREERUN
        uow.number_of_retries = 0
        uow.arguments = arguments
        uow.db_id = self.uow_dao.insert(uow)

        msg = 'Created: UOW {0} for {1}@{2}.' \
              .format(uow.db_id, freerun_entry.schedulable_name, timeperiod)
        self._log_message(INFO, freerun_entry, msg)
        return uow

    def _publish_uow(self, freerun_entry, uow):
        self.mq_transmitter.publish_freerun_uow(freerun_entry, uow)
        msg = 'Published: UOW {0} for {1}.'.format(uow.db_id, freerun_entry.schedulable_name)
        self._log_message(INFO, freerun_entry, msg)

    def _find_flow_uow(self, flow_request):
        """ most of the FreerunProcessEntries for workflows are run-time only
            i.e. that they disappear after the Scheduler restart, unlike persistent UOW
            hence, we have to try to fetch UOW associated with workflow+step+timeperiod """
        if not flow_request:
            return None
        try:
            uow = self.uow_dao.get_by_params(process_name=flow_request.schedulable_name,
                                             timeperiod=flow_request.timeperiod,
                                             start_id=0,
                                             end_id=0)
        except:
            uow = None
        return uow

    def _reset_flow_uow(self, freerun_entry, uow, flow_request):
        """ there can be multiple freeruns for a single combination of workflow+step+timeperiod
            hence, we have to *recycle* finished UOW """
        process_entry = context.process_context[freerun_entry.process_name]
        arguments = process_entry.arguments
        arguments.update(freerun_entry.arguments)
        arguments.update(flow_request.arguments)

        uow.created_at = datetime.utcnow()
        uow.submitted_at = datetime.utcnow()
        uow.state = unit_of_work.STATE_REQUESTED
        uow.unit_of_work_type = unit_of_work.TYPE_FREERUN
        uow.number_of_retries = 0
        uow.arguments = arguments

    def insert_and_publish_uow(self, freerun_entry, flow_request=None, reset_uow=False):
        try:
            uow = self._insert_uow(freerun_entry, flow_request)
        except DuplicateKeyError as e:
            msg = 'Duplication of UOW found for {0}. Error msg: {1}'.format(freerun_entry.schedulable_name, e)
            self._log_message(WARNING, freerun_entry, msg)
            uow = self.uow_dao.recover_from_duplicatekeyerror(e)

        if flow_request and reset_uow:
            self._reset_flow_uow(freerun_entry, uow, flow_request)

        if uow is not None:
            # publish the created/caught up unit_of_work
            self._publish_uow(freerun_entry, uow)
            freerun_entry.related_unit_of_work = uow.db_id
            self.sfe_dao.update(freerun_entry)
        else:
            msg = 'PERSISTENT TIER ERROR! Unable to locate UOW for {0}' \
                  .format(freerun_entry.schedulable_name)
            self._log_message(WARNING, freerun_entry, msg)

    def _process_state_embryo(self, freerun_entry, flow_request=None):
        """ method creates unit_of_work and associates it with the FreerunProcessEntry """
        self.insert_and_publish_uow(freerun_entry, flow_request)

    def _process_state_in_progress(self, freerun_entry, uow):
        """ method that takes care of processing unit_of_work records in STATE_REQUESTED or STATE_IN_PROGRESS states"""
        self._publish_uow(freerun_entry, uow)

    def _process_terminal_state(self, freerun_entry, uow, flow_request=None):
        """ method that takes care of processing unit_of_work records in
            STATE_PROCESSED, STATE_NOOP, STATE_INVALID, STATE_CANCELED states"""
        msg = 'UOW for {0} found in state {1}.'.format(freerun_entry.schedulable_name, uow.state)
        self._log_message(INFO, freerun_entry, msg)
        self.insert_and_publish_uow(freerun_entry, flow_request, reset_uow=True)

    def manage_schedulable(self, freerun_entry, flow_request=None):
        """ method main duty - is to _avoid_ publishing another unit_of_work, if previous was not yet processed
            In case the Scheduler sees that the unit_of_work is pending,
            it will issue new WorkerMqRequest for the same UOW """

        assert isinstance(freerun_entry, FreerunProcessEntry)
        if freerun_entry.related_unit_of_work is None:
            uow = self.uow_dao.get_one(freerun_entry.related_unit_of_work)
        else:
            uow = self._find_flow_uow(flow_request)

        try:
            if uow is None:
                self._process_state_embryo(freerun_entry, flow_request)

            elif uow.is_requested or uow.is_in_progress:
                self._process_state_in_progress(freerun_entry, uow)

            elif uow.is_finished or uow.is_invalid:
                self._process_terminal_state(freerun_entry, uow, flow_request)

            else:
                msg = 'Unknown state {0} of the UOW {1}'.format(uow.state, uow.db_id)
                self._log_message(ERROR, freerun_entry, msg)

        except LookupError as e:
            msg = 'Lookup issue for schedulable: {0} in timeperiod {1}, because of: {2}' \
                  .format(freerun_entry.db_id, uow.timeperiod, e)
            self._log_message(WARNING, freerun_entry, msg)

    def cancel_uow(self, freerun_entry):
        uow_id = freerun_entry.related_unit_of_work
        if uow_id is None:
            msg = 'cancel_uow: no related UOW for {0}'.format(freerun_entry.schedulable_name)
        else:
            uow = self.uow_dao.get_one(uow_id)
            uow.state = unit_of_work.STATE_CANCELED
            self.uow_dao.update(uow)
            msg = 'cancel_uow: canceled UOW {0} for {1}'.format(uow_id, freerun_entry.schedulable_name)
        self._log_message(INFO, freerun_entry, msg)



__author__ = 'Bohdan Mushkevych'

from odm.document import BaseDocument
from odm.fields import ObjectIdField, DictField, ListField, DateTimeField

PARENT_OBJECT_ID = 'parent_object_id'
LOG = 'log'
DETAILS = 'details'
CREATED_AT = 'created_at'


class LogRecording(BaseDocument):
    """ Persistent model: represents log/processing details from some activity,
        such as the Unit Of Work execution """
    db_id = ObjectIdField('_id', null=True)
    parent_object_id = ObjectIdField(PARENT_OBJECT_ID)
    log = ListField(LOG)
    details = DictField(DETAILS)
    created_at = DateTimeField(CREATED_AT)  # Field used by TTL mechanism

    @BaseDocument.key.getter
    def key(self):
        return self.parent_object_id

    @key.setter
    def key(self, value):
        self.parent_object_id = value



__author__ = 'Bohdan Mushkevych'

from datetime import datetime

from synergy.scheduler.process_hierarchy import ProcessHierarchy
from synergy.scheduler.tree_node import TreeNode, RootNode
from synergy.conf import settings
from synergy.system import time_helper
from synergy.system.time_helper import cast_to_time_qualifier


# number of times a Job can fail before it is considered STATE_SKIPPED.
MAX_NUMBER_OF_FAILURES = 3


class AbstractTree(object):
    """ MixIn to handle subscription for various tree events """

    def __init__(self):
        self.dependent_on = []

    def register_dependent_on(self, tree):
        """registering tree that we are dependent on.
        example: horizontal client should not be finalized until we have finalized vertical site for the same period"""
        self.dependent_on.append(tree)

    def unregister_dependent_on(self, tree):
        """unregistering tree that we are dependent on"""
        if tree in self.dependent_on:
            self.dependent_on.remove(tree)


class MultiLevelTree(AbstractTree):
    """ Multi-level Tree, suited to host both single Process Entry
        or multiple hierarchy-organized Process Entries """

    def __init__(self, process_names, timetable, tree_name=None, mx_name=None, mx_page=None):
        """
        :param tree_name <optional>: full name of the tree. used as an identifier
        :param mx_name <optional>: is used by MX only as visual vertical name
        :param mx_page <optional>: is used by MX only as anchor to specific page
        """
        super(MultiLevelTree, self).__init__()
        self.process_hierarchy = ProcessHierarchy(*process_names)

        self.build_timeperiod = None
        self.validation_timestamp = None
        self.timetable = timetable
        self.tree_name = tree_name
        self.mx_name = mx_name
        self.mx_page = mx_page
        self.root = RootNode(self)

    def __contains__(self, value):
        """
        :param value: process name
        :return: True if a process_entry with the given name is registered in this hierarchy; False otherwise
        """
        return value in self.process_hierarchy

    def _get_next_parent_node(self, parent):
        """ Used by _get_next_child_node, this method is called to find next possible parent.
            For example if timeperiod 2011010200 has all children processed, but is not yet processed itself
            then it makes sense to look in 2011010300 for hourly nodes """
        grandparent = parent.parent
        if grandparent is None:
            # here, we work at yearly/linear level
            return None

        parent_siblings = list(grandparent.children)
        sorted_keys = sorted(parent_siblings)
        index = sorted_keys.index(parent.timeperiod)
        if index + 1 >= len(sorted_keys):
            return None
        else:
            return grandparent.children[sorted_keys[index + 1]]

    def _get_next_child_node(self, parent):
        """
            Iterates among children of the given parent and looks for a suitable node to process
            In case given parent has no suitable nodes, a younger parent will be found
            and the logic will be repeated for him
        """
        children_keys = list(parent.children)
        sorted_keys = sorted(children_keys)
        for key in sorted_keys:
            node = parent.children[key]
            if node.job_record is None:
                self.timetable.assign_job_record(node)
                return node
            elif self.should_skip_tree_node(node):
                continue
            elif node.job_record.is_active:
                return node

        # special case, when all children of the parent node are not suitable for processing
        new_parent = self._get_next_parent_node(parent)
        if new_parent is not None:
            # in case all nodes are processed or blocked - look for next valid parent node
            return self._get_next_child_node(new_parent)
        else:
            # if all valid parents are exploited - return current node
            process_name = parent.children[sorted_keys[0]].process_name
            time_qualifier = parent.children[sorted_keys[0]].time_qualifier
            actual_timeperiod = time_helper.actual_timeperiod(time_qualifier)
            return self.get_node(process_name, actual_timeperiod)

    def _get_node(self, time_qualifier, timeperiod):
        """
        Method retrieves a tree node identified by time_qualifier and timeperiod
        In case intermittent or target nodes does not exist - method will request their creation
        :param time_qualifier: identifies the tree level
        :param timeperiod: identifies the target node timeperiod
        :return: requested node; type <AbstractNode>
        """
        hierarchy_entry = self.process_hierarchy.get_by_qualifier(time_qualifier)
        if hierarchy_entry.parent:
            parent_time_qualifier = hierarchy_entry.parent.process_entry.time_qualifier
            parent_timeperiod = hierarchy_entry.parent.cast_timeperiod(timeperiod)
            parent = self._get_node(parent_time_qualifier, parent_timeperiod)
        else:
            parent = self.root

        node = parent.children.get(timeperiod)
        if node is None:
            node = TreeNode(self, parent, hierarchy_entry.process_entry.process_name, timeperiod, None)
            parent.children[timeperiod] = node

        return node

    def _get_next_node(self, time_qualifier):
        """ Method goes to the top of the tree and traverses from
            there in search of the next suitable node for processing
            to the level defined by the given time_qualifier
            :param time_qualifier: defines target level of the tree
            :return: located node; type <TreeNode> """
        hierarchy_entry = self.process_hierarchy.get_by_qualifier(time_qualifier)
        if hierarchy_entry.parent:
            parent_time_qualifier = hierarchy_entry.parent.process_entry.time_qualifier
            parent = self._get_next_node(parent_time_qualifier)
        else:
            parent = self.root

        return self._get_next_child_node(parent)

    def should_skip_tree_node(self, node):
        """ :return True: in case the node should be _skipped_ and not included into processing """
        # case 1: node processing is complete
        if node.job_record.is_finished:
            return True

        # case 2: this is a bottom-level leaf node. retry this node for MAX_NUMBER_OF_FAILURES
        if node.process_name == self.process_hierarchy.bottom_process.process_name:
            if len(node.children) == 0:
                # no children - this is a leaf
                return node.job_record.number_of_failures > MAX_NUMBER_OF_FAILURES

        # case 3: here we process process_daily, process_monthly and process_yearly that have children
        # iterate thru children and check if all of them are in STATE_SKIPPED (i.e. no data for parent to process)
        # if any is still in processing (i.e. has produced some data) - then we can not skip parent of the child node
        # case 3': consider parent as worth processing (i.e. do not skip) if child's job_record is None
        all_children_spoiled = True
        for key, child in node.children.items():
            if child.job_record is None or \
                    (child.job_record.number_of_failures <= MAX_NUMBER_OF_FAILURES
                     and not child.job_record.is_skipped):
                all_children_spoiled = False
                break
        return all_children_spoiled

    def build_tree(self, rebuild=False):
        """ method builds tree by iterating from the synergy_start_timeperiod to the current time
            and inserting corresponding nodes """

        time_qualifier = self.process_hierarchy.bottom_process.time_qualifier
        process_name = self.process_hierarchy.bottom_process.process_name
        if rebuild or self.build_timeperiod is None:
            timeperiod = settings.settings['synergy_start_timeperiod']
        else:
            timeperiod = self.build_timeperiod

        timeperiod = cast_to_time_qualifier(time_qualifier, timeperiod)
        actual_timeperiod = time_helper.actual_timeperiod(time_qualifier)

        while actual_timeperiod >= timeperiod:
            self.get_node(process_name, timeperiod)
            timeperiod = time_helper.increment_timeperiod(time_qualifier, timeperiod)

        self.build_timeperiod = actual_timeperiod

    def get_next_node(self, process_name):
        """ :return: <AbstractTreeNode> next node to process by a process with process_name """
        if process_name not in self.process_hierarchy:
            raise ValueError('unable to compute the next_node due to unknown process: {0}'.format(process_name))

        time_qualifier = self.process_hierarchy[process_name].process_entry.time_qualifier
        return self._get_next_node(time_qualifier)

    def update_node(self, job_record):
        """ Updates job record property for a tree node associated with the given Job """
        if job_record.process_name not in self.process_hierarchy:
            raise ValueError('unable to update the node due to unknown process: {0}'.format(job_record.process_name))

        time_qualifier = self.process_hierarchy[job_record.process_name].process_entry.time_qualifier
        node = self._get_node(time_qualifier, job_record.timeperiod)
        node.job_record = job_record

    def get_node(self, process_name, timeperiod):
        """ Method retrieves a tree node identified by the time_qualifier and the timeperiod """
        if process_name not in self.process_hierarchy:
            raise ValueError('unable to retrieve the node due to unknown process: {0}'.format(process_name))

        time_qualifier = self.process_hierarchy[process_name].process_entry.time_qualifier
        return self._get_node(time_qualifier, timeperiod)

    def validate(self):
        """ method starts validation of the tree.
            @see TreeNode.validate """
        for timeperiod, child in self.root.children.items():
            child.validate()
        self.validation_timestamp = datetime.utcnow()



""" Module contains functions that are used to format, transform or classify date-time presented in a synergy format """

__author__ = 'Bohdan Mushkevych'

import calendar

from datetime import datetime, timedelta
from synergy.system.time_qualifier import *

SYNERGY_SESSION_PATTERN = '%Y%m%d%H%M%S'
SYNERGY_HOURLY_PATTERN = '%Y%m%d%H'
SYNERGY_DAILY_PATTERN = '%Y%m%d00'
SYNERGY_MONTHLY_PATTERN = '%Y%m0000'
SYNERGY_YEARLY_PATTERN = '%Y000000'


def define_pattern(timeperiod):
    if len(timeperiod) > 10:
        return SYNERGY_SESSION_PATTERN

    has_year = False
    has_month = False
    has_day = False
    has_hour = False

    for index in range(0, len(timeperiod)):
        if 0 <= index < 4 and timeperiod[index] != '0':
            has_year = True
        elif 4 <= index < 6 and timeperiod[index] != '0':
            has_month = True
        elif 6 <= index < 8 and timeperiod[index] != '0':
            has_day = True
        elif index >= 8 and timeperiod[index] != '0':
            has_hour = True

    pattern = ''
    if has_year:
        pattern += '%Y'
    else:
        pattern += '0000'
    if has_month:
        pattern += '%m'
    else:
        pattern += '00'
    if has_day:
        pattern += '%d'
    else:
        pattern += '00'
    if has_hour:
        pattern += '%H'
    else:
        pattern += '00'

    return pattern


def raw_to_session(timestamp):
    """ :param timestamp: <float> from time.time() output
    :return string in YYYYMMDDHHmmSS format"""
    t = datetime.utcfromtimestamp(timestamp)
    return t.strftime(SYNERGY_SESSION_PATTERN)


def session_to_hour(timestamp):
    """:param timestamp: as string in YYYYMMDDHHmmSS format
    :return string in YYYYMMDDHH format"""
    t = datetime.strptime(timestamp, SYNERGY_SESSION_PATTERN)
    return t.strftime(SYNERGY_HOURLY_PATTERN)


def hour_to_day(timeperiod):
    """:param timeperiod: as string in YYYYMMDDHH format
    :return string in YYYYMMDD00 format"""
    t = datetime.strptime(timeperiod, SYNERGY_HOURLY_PATTERN)
    return t.strftime(SYNERGY_DAILY_PATTERN)


def day_to_month(timeperiod):
    """:param timeperiod: as string in YYYYMMDD00 format
    :return string in YYYYMM0000 format"""
    t = datetime.strptime(timeperiod, SYNERGY_DAILY_PATTERN)
    return t.strftime(SYNERGY_MONTHLY_PATTERN)


def month_to_year(timeperiod):
    """:param timeperiod: as string in YYYYMM0000 format
    :return string in YYYY000000 format"""
    t = datetime.strptime(timeperiod, SYNERGY_MONTHLY_PATTERN)
    return t.strftime(SYNERGY_YEARLY_PATTERN)


def actual_timeperiod(time_qualifier):
    """ method receives process' time qualifier (hourly/daily/monthly/yearly)
    :return string representing current datetime (utc now) in requested Synergy Date format """
    return datetime_to_synergy(time_qualifier, datetime.utcnow())


def increment_timeperiod(time_qualifier, timeperiod, delta=1):
    """ method performs simple increment/decrement of the timeperiods
    For instance: 2010010119 with delta=1 -> 2010010120
    Or 2010010000 with delta=-1 -> 2009120000, etc"""

    pattern = define_pattern(timeperiod)
    t = datetime.strptime(timeperiod, pattern)

    if time_qualifier == QUALIFIER_HOURLY:
        t = t + timedelta(hours=delta)
        return t.strftime(SYNERGY_HOURLY_PATTERN)

    elif time_qualifier == QUALIFIER_DAILY:
        t = t + timedelta(days=delta)
        return t.strftime(SYNERGY_DAILY_PATTERN)

    elif time_qualifier == QUALIFIER_MONTHLY:
        yearly_increment = abs(delta) // 12
        yearly_increment = yearly_increment if delta >= 0 else -yearly_increment
        monthly_increment = delta - yearly_increment * 12

        if t.month + monthly_increment > 12:
            new_month = t.month + monthly_increment - 12
            new_year = t.year + yearly_increment + 1
            t = t.replace(year=new_year, month=new_month)
        elif t.month + monthly_increment < 1:
            new_month = t.month + monthly_increment + 12
            new_year = t.year + yearly_increment - 1
            t = t.replace(year=new_year, month=new_month)
        else:
            t = t.replace(year=t.year + yearly_increment, month=t.month + monthly_increment)
        return t.strftime(SYNERGY_MONTHLY_PATTERN)

    elif time_qualifier == QUALIFIER_YEARLY:
        t = t.replace(year=t.year + delta)
        return t.strftime(SYNERGY_YEARLY_PATTERN)
    else:
        raise ValueError('unknown time qualifier: {0}'.format(time_qualifier))


def cast_to_time_qualifier(time_qualifier, timeperiod):
    """ method casts given timeperiod accordingly to time qualifier.
    For example, will cast session time format of 20100101193412 to 2010010119 with QUALIFIER_HOURLY """

    if time_qualifier == QUALIFIER_HOURLY:
        date_format = SYNERGY_HOURLY_PATTERN
    elif time_qualifier == QUALIFIER_DAILY:
        date_format = SYNERGY_DAILY_PATTERN
    elif time_qualifier == QUALIFIER_MONTHLY:
        date_format = SYNERGY_MONTHLY_PATTERN
    elif time_qualifier == QUALIFIER_YEARLY:
        date_format = SYNERGY_YEARLY_PATTERN
    else:
        raise ValueError('unknown time qualifier: {0}'.format(time_qualifier))

    pattern = define_pattern(timeperiod)
    t = datetime.strptime(timeperiod, pattern)
    return t.strftime(date_format)


def datetime_to_synergy(time_qualifier, dt):
    """ method parses datetime and returns Synergy Date"""
    if time_qualifier == QUALIFIER_HOURLY:
        date_format = SYNERGY_HOURLY_PATTERN
    elif time_qualifier == QUALIFIER_DAILY:
        date_format = SYNERGY_DAILY_PATTERN
    elif time_qualifier == QUALIFIER_MONTHLY:
        date_format = SYNERGY_MONTHLY_PATTERN
    elif time_qualifier == QUALIFIER_YEARLY:
        date_format = SYNERGY_YEARLY_PATTERN
    elif time_qualifier == QUALIFIER_REAL_TIME:
        date_format = SYNERGY_SESSION_PATTERN
    else:
        raise ValueError('unknown time qualifier: {0}'.format(time_qualifier))
    return dt.strftime(date_format)


def synergy_to_datetime(time_qualifier, timeperiod):
    """ method receives timeperiod in Synergy format YYYYMMDDHH and convert it to UTC _naive_ datetime"""
    if time_qualifier == QUALIFIER_HOURLY:
        date_format = SYNERGY_HOURLY_PATTERN
    elif time_qualifier == QUALIFIER_DAILY:
        date_format = SYNERGY_DAILY_PATTERN
    elif time_qualifier == QUALIFIER_MONTHLY:
        date_format = SYNERGY_MONTHLY_PATTERN
    elif time_qualifier == QUALIFIER_YEARLY:
        date_format = SYNERGY_YEARLY_PATTERN
    elif time_qualifier == QUALIFIER_REAL_TIME:
        date_format = SYNERGY_SESSION_PATTERN
    else:
        raise ValueError('unknown time qualifier: {0}'.format(time_qualifier))
    return datetime.strptime(timeperiod, date_format).replace(tzinfo=None)


def session_to_epoch(timestamp):
    """ converts Synergy Timestamp for session to UTC zone seconds since epoch """
    utc_timetuple = datetime.strptime(timestamp, SYNERGY_SESSION_PATTERN).replace(tzinfo=None).utctimetuple()
    return calendar.timegm(utc_timetuple)


def tokenize_timeperiod(timeperiod):
    """
    method breaks given timeperiod into 4 parts: (year, month, day, hour)
    for instance: daily   2015031400 -> ('2013', '03', '14', '00')
                  hourly  2015031413 -> ('2013', '03', '14', '13')
                  monthly 2015030000 -> ('2013', '03', '00', '00')
                  yearly  2015000000 -> ('2015', '00', '00', '00')
    :return: tuple of four values
    """
    assert len(timeperiod) == 10, 'timeperiod {0} does not match accepted format YYYYMMDDHH'.format(timeperiod)
    return timeperiod[:4], timeperiod[4: 6], timeperiod[6: 8], timeperiod[8:],


if __name__ == '__main__':
    pass



__author__ = 'Bohdan Mushkevych'

from threading import RLock

from bson import ObjectId
from six import string_types

from synergy.db.manager import ds_manager
from synergy.system.decorator import thread_safe


def build_db_query(fields_names, field_values):
    """ method builds query dictionary by zipping together DB field names with the field values """
    if isinstance(field_values, string_types):
        field_values = [field_values]

    if len(fields_names) != len(field_values):
        raise ValueError('Error: unable to build a primary key query due '
                         'to mismatch in number of fields {0} vs {1}'
                         .format(len(fields_names), len(field_values)))

    query = dict()
    for k, v in zip(fields_names, field_values):
        query[k] = v
    return query


class BaseDao(object):
    """ Thread-safe base Data Access Object """

    def __init__(self, logger, model_class, primary_key, collection_name):
        super(BaseDao, self).__init__()
        self.logger = logger
        self.model_klass = model_class
        self.primary_key = primary_key
        self.collection_name = collection_name

        self.lock = RLock()
        self.ds = ds_manager.ds_factory(logger)

    @thread_safe
    def get_one(self, key):
        """ method finds single record base on the given primary key and returns it to the caller"""
        query = build_db_query(self.primary_key, key)
        collection = self.ds.connection(self.collection_name)

        document = collection.find_one(query)
        if document is None:
            raise LookupError('{0} with key {1} was not found'.format(self.model_klass.__name__, query))
        return self.model_klass.from_json(document)

    @thread_safe
    def run_query(self, query):
        """ method runs query on a specified collection and return a list of filtered Model records """
        collection = self.ds.connection(self.collection_name)

        cursor = collection.find(query)
        if cursor.count() == 0:
            raise LookupError('Collection {0} has no {1} records'
                              .format(self.collection_name, self.model_klass.__name__))
        return [self.model_klass.from_json(entry) for entry in cursor]

    @thread_safe
    def get_all(self):
        return self.run_query({})

    @thread_safe
    def update(self, instance):
        """ this is an upsert method: replaces or creates the DB representation of the model instance """
        assert isinstance(instance, self.model_klass)
        if instance.db_id:
            query = {'_id': ObjectId(instance.db_id)}
        else:
            query = build_db_query(self.primary_key, instance.key)
        self.ds.update(self.collection_name, query, instance)
        return instance.db_id

    @thread_safe
    def remove(self, key):
        query = build_db_query(self.primary_key, key)
        collection = self.ds.connection(self.collection_name)
        collection.delete_one(query)



__author__ = 'Bohdan Mushkevych'

from collections import OrderedDict
from threading import RLock

from werkzeug.utils import cached_property

from synergy.db.dao import job_dao
from synergy.db.dao.job_dao import JobDao
from synergy.db.dao import unit_of_work_dao
from synergy.db.dao.unit_of_work_dao import UnitOfWorkDao
from synergy.scheduler.scheduler_constants import COLLECTION_JOB_YEARLY, \
    COLLECTION_JOB_MONTHLY, COLLECTION_JOB_DAILY, COLLECTION_JOB_HOURLY
from synergy.system.decorator import thread_safe
from synergy.system import time_helper
from synergy.system.time_qualifier import QUALIFIER_DAILY, QUALIFIER_MONTHLY, QUALIFIER_YEARLY
from synergy.mx.base_request_handler import BaseRequestHandler, valid_action_request


class DashboardHandler(BaseRequestHandler):
    def __init__(self, request, **values):
        super(DashboardHandler, self).__init__(request, **values)

        self.time_window = self.request.args.get('time_window')
        self.is_include_running = self.request.args.get('include_running') == 'on'
        self.is_include_processed = self.request.args.get('include_processed') == 'on'
        self.is_include_noop = self.request.args.get('include_noop') == 'on'
        self.is_include_failed = self.request.args.get('include_failed') == 'on'
        self.is_include_disabled = self.request.args.get('include_disabled') == 'on'
        self.is_request_valid = bool(self.time_window)

        if self.is_request_valid:
            actual_timeperiod = time_helper.actual_timeperiod(QUALIFIER_DAILY)
            delta = int(self.time_window)
            self.query_start_timeperiod = time_helper.increment_timeperiod(QUALIFIER_DAILY, actual_timeperiod, -delta)

    @cached_property
    @valid_action_request
    def managed(self):
        processor = ManagedStatements(self.logger, self.scheduler.managed_handlers)
        selection = processor.retrieve_records(self.query_start_timeperiod, self.is_include_running,
                                               self.is_include_processed, self.is_include_noop, self.is_include_failed,
                                               self.is_include_disabled)
        return OrderedDict(sorted(selection.items()))

    @cached_property
    @valid_action_request
    def freeruns(self):
        processor = FreerunStatements(self.logger, self.scheduler.freerun_handlers)
        selection = processor.retrieve_records(self.query_start_timeperiod, self.is_include_running,
                                               self.is_include_processed, self.is_include_noop, self.is_include_failed,
                                               self.is_include_disabled)
        return OrderedDict(sorted(selection.items()))


class ManagedStatements(object):
    def __init__(self, logger, managed_handlers):
        self.lock = RLock()
        self.logger = logger
        self.managed_handlers = managed_handlers
        self.job_dao = JobDao(self.logger)

    @thread_safe
    def retrieve_records(self, timeperiod, include_running,
                         include_processed, include_noop, include_failed, include_disabled):
        """ method looks for suitable job records in all Job collections and returns them as a dict"""
        resp = dict()
        resp.update(self._search_by_level(COLLECTION_JOB_HOURLY, timeperiod, include_running,
                                          include_processed, include_noop, include_failed, include_disabled))
        resp.update(self._search_by_level(COLLECTION_JOB_DAILY, timeperiod, include_running,
                                          include_processed, include_noop, include_failed, include_disabled))

        timeperiod = time_helper.cast_to_time_qualifier(QUALIFIER_MONTHLY, timeperiod)
        resp.update(self._search_by_level(COLLECTION_JOB_MONTHLY, timeperiod, include_running,
                                          include_processed, include_noop, include_failed, include_disabled))

        timeperiod = time_helper.cast_to_time_qualifier(QUALIFIER_YEARLY, timeperiod)
        resp.update(self._search_by_level(COLLECTION_JOB_YEARLY, timeperiod, include_running,
                                          include_processed, include_noop, include_failed, include_disabled))
        return resp

    @thread_safe
    def _search_by_level(self, collection_name, timeperiod, include_running,
                         include_processed, include_noop, include_failed, include_disabled):
        resp = dict()
        try:
            query = job_dao.QUERY_GET_LIKE_TIMEPERIOD(timeperiod, include_running,
                                                      include_processed, include_noop, include_failed)
            records_list = self.job_dao.run_query(collection_name, query)
            if len(records_list) == 0:
                self.logger.warning('MX: no Job Records found in {0} since {1}.'.format(collection_name, timeperiod))

            for job_record in records_list:
                if job_record.process_name not in self.managed_handlers:
                    continue

                thread_handler = self.managed_handlers[job_record.process_name]
                if not include_disabled and not thread_handler.process_entry.is_on:
                    continue

                resp[job_record.key] = job_record.document
        except Exception as e:
            self.logger.error('MX Dashboard ManagedStatements error: {0}'.format(e))
        return resp


class FreerunStatements(object):
    def __init__(self, logger, freerun_handlers):
        self.lock = RLock()
        self.logger = logger
        self.freerun_handlers = freerun_handlers
        self.uow_dao = UnitOfWorkDao(self.logger)

    @thread_safe
    def retrieve_records(self, timeperiod, include_running,
                         include_processed, include_noop, include_failed, include_disabled):
        """ method looks for suitable UOW records and returns them as a dict"""
        resp = dict()
        try:
            query = unit_of_work_dao.QUERY_GET_FREERUN_SINCE(timeperiod, include_running,
                                                             include_processed, include_noop, include_failed)
            records_list = self.uow_dao.run_query(query)
            if len(records_list) == 0:
                self.logger.warning('MX: no Freerun UOW records found since {0}.'.format(timeperiod))

            for uow_record in records_list:
                if uow_record.process_name not in self.freerun_handlers:
                    continue

                thread_handler = self.freerun_handlers[uow_record.process_name]
                if not include_disabled and not thread_handler.process_entry.is_on:
                    continue

                resp[uow_record.key] = uow_record.document
        except Exception as e:
            self.logger.error('MX Dashboard FreerunStatements error: {0}'.format(e))
        return resp



__author__ = 'Bohdan Mushkevych'

from logging import ERROR, INFO

from synergy.db.model import job
from synergy.scheduler.scheduler_constants import STATE_MACHINE_DISCRETE
from synergy.scheduler.abstract_state_machine import AbstractStateMachine
from synergy.system import time_helper
from synergy.conf import context


class StateMachineDiscrete(AbstractStateMachine):
    """ State Machine of 5 states is expected to spawn one UOW per job/timeperiod
        Job timeperiods boundaries are meant to be discrete/fixed
        in comparison to StateMachineContinuous this one does not transfer to STATE_FINAL_RUN"""

    def __init__(self, logger, timetable):
        super(StateMachineDiscrete, self).__init__(logger, timetable, name=STATE_MACHINE_DISCRETE)

    @property
    def run_on_active_timeperiod(self):
        """ :return: False, since there should be only 1 run for given timeperiod """
        return False

    def notify(self, uow):
        tree = self.timetable.get_tree(uow.process_name)
        node = tree.get_node(uow.process_name, uow.timeperiod)
        job_record = node.job_record

        if not job_record.is_in_progress:
            self.logger.info('Suppressing state change for Job {0}@{1}, since it is not in STATE_IN_PROGRESS'
                             .format(uow.process_name, uow.timeperiod))
            return

        time_qualifier = context.process_context[uow.process_name].time_qualifier
        actual_timeperiod = time_helper.actual_timeperiod(time_qualifier)
        is_job_finalizable = self.timetable.is_job_record_finalizable(job_record)

        if uow.timeperiod < actual_timeperiod and is_job_finalizable is True:
            self.__process_finalizable_job(job_record, uow)
            self.mq_transmitter.publish_job_status(job_record)

        elif uow.timeperiod >= actual_timeperiod:
            self.logger.info('Suppressing state change for Job {0}@{1}, '
                             'since the working timeperiod has not finished yet'
                             .format(uow.process_name, uow.timeperiod))

        elif not is_job_finalizable:
            self.logger.info('Suppressing state change for Job {0}@{1}, '
                             'since the job is not finalizable'.format(uow.process_name, uow.timeperiod))

    def __process_non_finalizable_job(self, job_record, uow):
        """ method handles given job_record based on the unit_of_work status
        Assumption: job_record is in STATE_IN_PROGRESS and is not yet finalizable """
        if uow.is_active:
            # Large Job processing takes more than 1 tick of Scheduler
            # Let the Large Job processing complete - do no updates to Scheduler records
            pass
        elif uow.is_finished:
            # create new uow to cover new inserts
            uow, is_duplicate = self.insert_and_publish_uow(job_record, 0, int(uow.end_id) + 1)
            self.update_job(job_record, uow, job.STATE_IN_PROGRESS)

    def __process_finalizable_job(self, job_record, uow):
        """ method handles given job_record based on the unit_of_work status
        Assumption: job_record is in STATE_IN_PROGRESS and is finalizable """
        if uow.is_active:
            # Job processing has not started yet
            # Let the processing complete - do no updates to Scheduler records
            msg = 'Suppressed new UOW creation for Job {0}@{1}; Job is in {2}; UOW is in {3}' \
                  .format(job_record.process_name, job_record.timeperiod, job_record.state, uow.state)
            self._log_message(INFO, job_record.process_name, job_record.timeperiod, msg)
        elif uow.is_processed:
            self.update_job(job_record, uow, job.STATE_PROCESSED)
        elif uow.is_noop:
            self.update_job(job_record, uow, job.STATE_NOOP)
        elif uow.is_canceled:
            self.update_job(job_record, uow, job.STATE_SKIPPED)
        elif uow.is_invalid:
            msg = 'Job {0}: UOW for {1}@{2} is in {3}; ' \
                  'relying on the Garbage Collector to transfer UOW into the STATE_CANCELED' \
                  .format(job_record.db_id, job_record.process_name, job_record.timeperiod, uow.state)
            self._log_message(INFO, job_record.process_name, job_record.timeperiod, msg)
        else:
            msg = 'Unknown state {0} of Job {1} for {2}@{3}' \
                  .format(uow.state, job_record.db_id, job_record.process_name, job_record.timeperiod)
            self._log_message(INFO, job_record.process_name, job_record.timeperiod, msg)

        timetable_tree = self.timetable.get_tree(job_record.process_name)
        timetable_tree.build_tree()

    def _process_state_embryo(self, job_record):
        """ method that takes care of processing job records in STATE_EMBRYO state"""
        uow, is_duplicate = self.insert_and_publish_uow(job_record, 0, 0)
        self.update_job(job_record, uow, job.STATE_IN_PROGRESS)

    def _process_state_in_progress(self, job_record):
        """ method that takes care of processing job records in STATE_IN_PROGRESS state """
        time_qualifier = context.process_context[job_record.process_name].time_qualifier
        actual_timeperiod = time_helper.actual_timeperiod(time_qualifier)
        is_job_finalizable = self.timetable.is_job_record_finalizable(job_record)
        uow = self.uow_dao.get_one(job_record.related_unit_of_work)

        if job_record.timeperiod == actual_timeperiod or is_job_finalizable is False:
            self.__process_non_finalizable_job(job_record, uow)

        elif job_record.timeperiod < actual_timeperiod and is_job_finalizable is True:
            self.__process_finalizable_job(job_record, uow)

        else:
            msg = 'Job {0} has timeperiod {1} from the future vs current timeperiod {2}' \
                  .format(job_record.db_id, job_record.timeperiod, actual_timeperiod)
            self._log_message(ERROR, job_record.process_name, job_record.timeperiod, msg)

    def _process_state_final_run(self, job_record):
        """method takes care of processing job records in STATE_FINAL_RUN state"""
        raise NotImplementedError('Method _process_state_final_run is not supported by {0}'
                                  .format(self.__class__.__name__))



from synergy.db.model.queue_context_entry import queue_context_entry
from synergy.scheduler.scheduler_constants import PROCESS_GC, TOKEN_GC, PROCESS_MX, TOKEN_WERKZEUG, EXCHANGE_UTILS, \
    PROCESS_SCHEDULER, TOKEN_SCHEDULER, QUEUE_UOW_STATUS, QUEUE_JOB_STATUS
from synergy.supervisor.supervisor_constants import PROCESS_SUPERVISOR, TOKEN_SUPERVISOR
from synergy.db.model.daemon_process_entry import daemon_context_entry

process_context = {
    PROCESS_MX: daemon_context_entry(
        process_name=PROCESS_MX,
        token=TOKEN_WERKZEUG,
        classname=''),

    PROCESS_GC: daemon_context_entry(
        process_name=PROCESS_GC,
        token=TOKEN_GC,
        classname=''),

    PROCESS_SCHEDULER: daemon_context_entry(
        process_name=PROCESS_SCHEDULER,
        classname='synergy.scheduler.synergy_scheduler.Scheduler.start',
        token=TOKEN_SCHEDULER),

    PROCESS_SUPERVISOR: daemon_context_entry(
        process_name=PROCESS_SUPERVISOR,
        classname='synergy.supervisor.synergy_supervisor.Supervisor.start',
        token=TOKEN_SUPERVISOR),
}

mq_queue_context = {
    QUEUE_UOW_STATUS: queue_context_entry(exchange=EXCHANGE_UTILS, queue_name=QUEUE_UOW_STATUS),
    QUEUE_JOB_STATUS: queue_context_entry(exchange=EXCHANGE_UTILS, queue_name=QUEUE_JOB_STATUS),
}

timetable_context = {
}



"""
 @author: Kalpesh Patel, Bohdan Mushkevych
 @date: April 6, 2010
 http://neopatel.blogspot.com/2010/04/rabbitmq-purging-queue.html
"""

import sys

from synergy.conf import context
from synergy.mq.flopsy import purge_mq_queue


if __name__ == '__main__':
    if len(sys.argv) < 2:
        print('Usage 1 - to purge specific queue:\n\tpython purge_queue.py queue_name\n')
        print('Usage 2 - to purge all Synergy queues:\n\tpython purge_queue.py all\n')
        sys.exit(0)

    if sys.argv[1] == 'all':
        print('Purging process-derived queues...')
        for process_name in context.process_context:
            queue_name = context.process_context[process_name].mq_queue
            if not queue_name:
                continue
            purge_mq_queue(queue_name)

        print('Purging custom queues...')
        for queue_name in context.mq_queue_context:
            if not queue_name:
                continue
            purge_mq_queue(queue_name)
    else:
        queue_name = sys.argv[1]
        purge_mq_queue(queue_name)
